/*
   dataopen.c

   Implementation of Data Transfer Process: initialization.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: dataopen.c,v 1.17 2008/04/28 05:37:53 shesmax Exp $ */

#include <sys/types.h>

#include <unistd.h>     /* for open, close */
#include <sys/stat.h>   /* for S_XXX macros */
#include <fcntl.h>      /* for fcntl codes */
#include <sys/socket.h> /* for socket */
#include <netinet/in.h>
#include <netinet/ip.h> /* fot TOS */
#include <sys/time.h>   /* gettimeofday */
#include <errno.h>      /* for errno */
#include <string.h>     /* for strerror */

#include "pi.h"         /* for PError */
#include "retcodes.h"
#include "dtp.h"        /* for nb_t */
#include "dataopen.h"
#include "chkport.h"
#include "sockets.h"    /* for AcceptSocket and other operations */
#include "log.h"        /* for LogEvent and codes */

/*
* Data connection handling
*/

/* Creates socket for the data connection. */
int CreateDataSocket( int* hDataSocket )
{
   socklen_t socklen;
   int x;

   /* if we have some socket open, close it (f.e. PASV has opened it) */
   if( *hDataSocket != INVALID_SOCKET )
      close( *hDataSocket );

   /* create the socket */
   *hDataSocket = socket( addrSrvrData->sa_family, SOCK_STREAM, 0 );
   if( *hDataSocket == INVALID_SOCKET )
      goto __error__;

   /* bind it with our inet address */
   x = 1;
   if( setsockopt( *hDataSocket, SOL_SOCKET, SO_REUSEADDR, &x, sizeof x ) == -1 )
      goto __error__;
   if( bind( *hDataSocket, addrSrvrData, SockAddrSize( addrSrvrData ) ) == -1 )
      goto __error__;
   /* get the address of the socket */
   socklen = SockAddrSize( addrSrvrData );
   if( getsockname( *hDataSocket, addrSrvrData, &socklen ) == -1 )
      goto __error__;

   if
   (
      (x = fcntl( *hDataSocket, F_GETFL )) == -1 ||
      fcntl( *hDataSocket, F_SETFL, x | O_NONBLOCK ) == -1
   )
      goto __error__;

   x = IPTOS_THROUGHPUT;
   if( setsockopt( *hDataSocket, IPPROTO_IP, IP_TOS, &x, sizeof x ) == -1 )
      LogEvent( LLOG_WARN, "IPTOS_THROUGHPUT: %s\n", strerror( errno ) );

   return RC_OK;

__error__:
   return RC_ERROR;
}

/* Closes the data socket */
#define NO_FURTHER_SENDS_AND_RECEIVES  2
static void CloseDataSocket( int* hDataSocket )
{
   if( *hDataSocket != INVALID_SOCKET )
   {
      shutdown( *hDataSocket, NO_FURTHER_SENDS_AND_RECEIVES );
      close( *hDataSocket );
      *hDataSocket = INVALID_SOCKET;
   }
}

/*
   Opens either passive or active data connection. Returns RC_TRUE if the
   data connection was opened.
*/
int OpenDataConnection( int* hDataSocket )
{
   if( bPassive )                                  /* Passive role */
   {
      if( SendReply( "125 Using existing data connection\r\n") == RC_OK )
      {
         int nSocket = AcceptSocket( *hDataSocket, addrPeerData );

         if( !CheckDataAddr( addrPeerCtrl, addrPeerData ) )
         {
            CloseDataSocket( hDataSocket );
            SendReply("425 Invalid address of data connection\r\n");
            return RC_ERROR;
         } 
         else if( nSocket != INVALID_SOCKET )
         {
            close( *hDataSocket );
            *hDataSocket = nSocket;
            return RC_OK;
         }
      }
   }
   else                                            /* Active role */
   {
      if( CreateDataSocket( hDataSocket ) == RC_OK &&
          SendReply( "150 Opening data connection\r\n" ) == RC_OK &&
          ConnectSocket( *hDataSocket, addrPeerData ) == RC_OK )
         return RC_OK;
   }

   /*
      Note: The error code 425 is also returned for PASV though it is
      not specified in RFC. This is done according to the information
      from the source code of Berkeley's FTP daemon.
      Following is the excerpt from ftpd.c
      *
      * Note: a response of 425 is not mentioned as a possible response to
      *  the PASV command in RFC959. However, it has been blessed as
      *  a legitimate response by Jon Postel in a telephone conversation
      *  with Rick Adams on 25 Jan 89.
   */
   PError( 425 );
   CloseDataSocket( hDataSocket );
   return RC_ERROR;
}

/* Closes data connection. */
void CloseDataConnection( int* hDataSocket )
{
   CloseDataSocket( hDataSocket );
   *hDataSocket = INVALID_SOCKET;
   bPassive = RC_FALSE;
}

/*
*  Sending / receiving
*/

int OpenFile( const char* szPath, int flags, mode_t mode, struct stat* st )
{
   int fd;

   fd = open( szPath, flags, mode );
   if( fd == INVALID_FILE_DESCRIPTOR || fstat( fd, st ) < 0 )
      PError( 550 );
   else if( !S_ISREG( st->st_mode ) )
      AccumReply( "550 Not a regular file\r\n" );
   else
      return fd;

   /* Something has failed */
   if( fd != INVALID_FILE_DESCRIPTOR )
      close( fd );
   return INVALID_FILE_DESCRIPTOR;
}

/* Logs transfer statistics */
void LogTransferStat
(
   const char* szPath, struct timeval* pstart,
   const char* szLabel, const char* szStatus
)
{
   struct timeval stop;
   nb_t nTransferSize;

   timerclear( &stop );
   (void) gettimeofday( &stop, NULL );
   if( timerisset( pstart ) && timerisset( &stop ) )
   {
      struct timeval diff;
      diff.tv_sec = stop.tv_sec - pstart->tv_sec;
      diff.tv_usec = stop.tv_usec - pstart->tv_usec;
      if( diff.tv_usec < 0 ) {
         diff.tv_usec += 1000000;
         diff.tv_sec--;
      }
      nTransferSize = 0;
      if( nRestartAt < nDtPassed )
         nTransferSize = nDtPassed - nRestartAt;
      LogEvent
      ( 
         LLOG_AUDIT, "%s: %s %d bytes %d.%06d secs, %s\n",
         szLabel, szPath, nTransferSize, diff.tv_sec, diff.tv_usec, szStatus
      );
   }
}
