/*
   dataretr.c

   Implementation of Data Transfer Process: retrieve.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: dataretr.c,v 1.12 2008/04/28 05:37:54 shesmax Exp $ */

#include <sys/types.h>

#include <unistd.h>     /* for mmap */
#include <sys/mman.h>   /* for mmap */
#include <fcntl.h>      /* for O_XXX constants */
#include <sys/time.h>   /* for gettimeofday */
#include <string.h>     /* for memchr, strcmp */
#include <sys/stat.h>   /* for struct stat */

#include "libra.h"      /* for nAsciiBufSize */
#include "pi.h"         /* for PError */
#include "retcodes.h"
#include "dtp.h"        /* for nb_t, TransferInProgress etc */
#include "dataopen.h"   /* for nDtPassed etc */
#include "sockets.h"    /* for INVALID_SOCKET, WriteSocket etc */
#include "xbuf.h"
#include "log.h"        /* for LogEvent and codes */
#include "md5.h"        /* for checksumming stuff */

/*
   Writes the content of file line by line to data connection.
   All newlines are replaced by CRLF.

   Parameters

   nSocket     Data connection socket.
   buf         Address of memory mapped part of file.
   nSend       Size of the part.

   Returns

   RC_OK          No errors.
   RC_ERROR       OS error occured.
   RC_TIMEDOUT    Sending through the data socket timed out.
   RC_ABORTED     Data transfer aborted by the user.
*/
static int SendAscii( int nSocket, const void* buf, int nSend )
{
   const char* begin;
   const char* end;
   const char* eol;
   int nResult;
   XBuf xbuf;
   int nBufSize;

   if( nSend >= nAsciiBufSize >> 1 )
      /* `>>' for the extreme case - newlines only */
      nBufSize = nAsciiBufSize;
   else
      nBufSize  = nSend << 1;
   if( CreateFixedXBuf( &xbuf, nBufSize ) == XBUF_OVERFLOW )
      return RC_ERROR;

   for( begin = (const char*)buf, end = begin + nSend;; )
   {
      if( !(eol = memchr( begin, '\n', end - begin )) )
         eol = end;
      nResult = PutSockXBuf( &xbuf, begin, eol-begin, FlushSockXBuf, nSocket );
      if( nResult != RC_OK )
         break;
      nResult = PutSockXBuf( &xbuf, "\r\n", 2, FlushSockXBuf, nSocket );
      if( nResult != RC_OK )
         break;
      if( (begin = eol + 1) >= end )
      {
         if( xbuf.nUsed > 0 )
            nResult = FlushSockXBuf( &xbuf, nSocket );
         break;
      }
   }

   DestroyXBuf( &xbuf );
   return nResult;
}

/*
   Sends the file to the remote host through the data connection.

   Parameters

   fd          File descriptor.
   sock        Output socket.
   SendFunc    The function that takes the address of the memory
               mapped chunk of file, prepares it for sending and sends
               through the data socket.  Sending in binary mode does not
               requires any conversion, hence we use WriteSocket.
               ASCII mode requeires to convert all newlines to CRLF.
   md5ctx      Context for MD5 checksumming (if necessary).
   nDtPassed   Transfer start offset rounded to nMaxMap.
               This value is a global progress indicator.  It is preserved
               valid (in quants of nMaxMap) during the whole operation and may
               be read asynchronously.
               (Kept in a global variable.)
   nPos        Transfer start offset remainder.  The real offset is
               nDtPassed+nPos.
               (Passed as a function parameter.)
   nDtTotal    Transfer end point (offset from the beginning of the file).
               (Kept in a global variable.)

   Returns

   RC_OK          No errors
   RC_ERROR       OS error occured.
   RC_TIMEDOUT    Sending through the data socket timed out.
   RC_ABORTED     Data transfer aborted by the user.
*/
static int SendFile
( 
   int fd, int sock, int (*SendFunc)( int, const void*, int ),
   int nPos, /* offset inside the buffer */
   struct MD5Context *md5ctx
)
{
   nb_t nSend;
   int nResult;
   char* buf;

   for( nResult = RC_OK;; )
   {
      if( nDtPassed >= nDtTotal )   /* The file is transferred */
         break;
      if( (nSend = nDtTotal - nDtPassed) > (unsigned)nMaxMap )
         nSend = nMaxMap;

      buf = (char*)mmap( 0, nSend, PROT_READ, MAP_PRIVATE, fd, nDtPassed );
      if( buf == (char*) -1 )
      {
         nResult = RC_ERROR;
         break;
      }

      if( ComputeChecksums() )
         MD5Update( md5ctx, (unsigned char*)buf + nPos, nSend - nPos );
      nResult = SendFunc( sock, buf + nPos, nSend - nPos );

      munmap( buf, nSend );
      if( nResult != RC_OK )
         break;
      nPos = 0;
      nDtPassed += nSend;
   }

   return nResult;
}

int DoRetrieve( int *hDataSocket, const char* szPath )
{
   int fd;
   struct stat st;

   /* No restarts for broken transfers in ASCII mode.
      They open a vulnerability to "denial of service" attacks.
    */
   if( nRestartAt && !strcmp( GetRepresentationType(), "Ascii" ) )
      return SendReply( "425 Restart is prohibited in ASCII mode\r\n" );

   fd = OpenFile( szPath, O_RDONLY, 0, &st );
   if( fd != INVALID_FILE_DESCRIPTOR )
   {
      if( OpenDataConnection( hDataSocket ) == RC_OK )
      {
         struct timeval start;
         struct MD5Context md5ctx;
         int nPos;
         int nResult;

         timerclear( &start );
         (void) gettimeofday( &start, NULL );
         LogEvent( LLOG_AUDIT, "Retr: %s\n", szPath );

         if( ComputeChecksums() )
            MD5Init( &md5ctx );
         nDtTotal = st.st_size;
         nPos = nRestartAt % nMaxMap;
         nDtPassed = nRestartAt - nPos;
         CancelTransferAbort();
         StartTransfer();

         nResult = SendFile
         ( 
            fd, *hDataSocket, !strcmp( GetRepresentationType(), "Ascii" ) ? SendAscii : WriteSocket,
            nPos, &md5ctx 
         );
         StopTransfer();
         CancelTransferAbort();

         switch( nResult )
         {
         case RC_ERROR:
            PError( 451 );
            break;
         case RC_TIMEDOUT:
            AccumReply( "550 Data transfer timeout\r\n" );
            break;
         case RC_ABORTED:
            LogTransferStat( szPath, &start, "Retr", "aborted" );
            AccumReply( "426 Data transfer aborted\r\n" );
            AccumReply( "226 File transferred\r\n" );
            break;
         default:
            LogTransferStat( szPath, &start, "Retr", "completed" );
            AccumReply( "226 File transferred." );
            if( ComputeChecksums() )
            {
               unsigned int i;
               unsigned char sChecksum[ 16 ];

               MD5Final( sChecksum, &md5ctx );
               AccumReply( " Checksum (MD5): " );
               for( i = 0; i < sizeof sChecksum; ++i )
                  AccumReplyF( "%02x", sChecksum[ i ] );
            }
            AccumReply( "\r\n" );
            break;
         }
         CloseDataConnection( hDataSocket );
      }
      close( fd );
   }

   /* reset restart point to be sure it won't affect next transfer */
   nRestartAt = 0;
   return SendReplies();
}
