/*
   datastor.c

   Implementation of Data Transfer Process: store.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: datastor.c,v 1.13 2002/12/14 23:16:36 shesmax Exp $ */

#include <sys/types.h>

#include <unistd.h>     /* for write, lseek */
#include <fcntl.h>      /* for O_XXX constants */
#include <sys/time.h>   /* for gettimeofday */
#include <string.h>     /* for memchr, strcmp */
#include <sys/stat.h>   /* for struct stat */
#include <errno.h>      /* for errno, EINTR */

#include "libra.h"      /* for ReadOnlyBehavior */
#include "pi.h"         /* for PError */
#include "retcodes.h"
#include "dtp.h"        /* for nb_t, TransferInProgress etc */
#include "dataopen.h"   /* for nDtPassed etc */
#include "sockets.h"    /* for INVALID_SOCKET, WriteSocket etc */
#include "xbuf.h"
#include "log.h"        /* for LogEvent and codes */
#include "md5.h"        /* for checksumming stuff */

/*
* Reception of the file from the remote host
*/

/*
   Writes the content of the buffer to the file fd.

   Parameters

   fd             File descriptor.
   buf            Received to store.
   nBytes         Number of bytes to write.
   md5ctx         Context for MD5 checksumming (if necessary).

   Returns

   RC_OK          No errors.
   RC_ERROR       OS level error.
*/
static int WriteBinary( int fd, const void* buf, int nBytes, struct MD5Context* md5ctx )
{
   int w;
   
   if( ComputeChecksums() )
      MD5Update( md5ctx, buf, nBytes );
   while( nBytes > 0 )
   {
      w = write( fd, buf, nBytes );
      if( w == -1 )
      {
         if( errno != EINTR ) 
            return RC_ERROR;
      }
      else
      {
         buf += w;
         nBytes -= w;
      }
   }
   return RC_OK;
}

/*
   Receives the binary file from the remote host through the data connection.

   Parameters

   fd             File descriptor.
   sock           Input socket.
   md5ctx         Context for MD5 checksumming (if necessary).

   Returns

   RC_OK          No errors.
   RC_ERROR       OS level error occured.
   RC_TIMEDOUT    Waiting for new data timed out.
   RC_ABORTED     Data transfer aborted by the user.
*/
static int RecvBinaryFile( int fd, int sock, struct MD5Context* md5ctx )
{
   char buf[ 4096 ];
   int nRead;
   int nBuf;
   
   if( nRestartAt )
   {
      if( lseek( fd, nRestartAt, SEEK_SET ) == -1 )
         return RC_ERROR;
   }

   for( ;; )
   {
      nRead = ReadSocket( sock, buf, sizeof buf );
      if( nRead < 0 )
         return RC_ERROR;
      if( nRead == 0 )
         return RC_OK;

      nBuf = WriteBinary( fd, buf, nRead, md5ctx );
      if( nBuf == RC_ERROR )
         return nBuf;
      nDtPassed += nRead;
   }
}

/*
   Writes the content of the buffer to the file fd.
   All CRLFs are replaced by newlines.

   Parameters

   fd             File descriptor.
   buf            Received data.
   nBytes         Number of bytes to write.
   md5ctx         Context for MD5 checksumming (if necessary).

   Returns

   0 || 1         Number of characters left to write in the buffer. No errors.
                  Note that it is assumed here RC_OK is greater or equal to 0.
   RC_ERROR       OS lelel error occured.
*/
static int WriteAscii( int fd, char* buf, int nBytes, struct MD5Context* md5ctx )
{
   char* bufp = buf;
   char* end = buf + nBytes;
   char* bol;
   char* eol;

   for( bol = buf; bol < end; bol = eol + 1 )
   {
      int n;
      
      eol = memchr( bol, '\r', end - bol  );
      if( !eol )
         eol = end;
      else if( eol < end - 1 && eol[ 1 ] != '\n' )
         continue;   /* write a standalone \r on next turn */

      /* Either \r\n or end of buffer or \r at the end of buffer */
      n = eol - bufp;
      if( n > 0 )
         if( WriteBinary( fd, bufp, n, md5ctx ) == RC_ERROR )
            return RC_ERROR;

      if( eol == end - 1 )
      {
         /* Transport layer splitted the stream right in the
            middle of \r\n sequence */
         *buf = '\r';
         return 1;
      }
      bufp = eol + 1;
   }

   return 0;
}

/*
   Receives an ASCII file from the remote host through the data connection.

   Parameters

   fd             File descriptor
   sock           Input socket.
   md5ctx         Data for MD5 digest.

   Returns

   RC_OK          No errors
   RC_ERROR       OS error occured.
   RC_TIMEDOUT    Waiting for new data timed out.
   RC_ABORTED     Data transfer aborted by the user.
*/
static int RecvAsciiFile( int fd, int sock, struct MD5Context* md5ctx )
{
   char buf[ 4096 ];
   int nRead;
   int nBuf = 0;
   
   for( ;; )
   {
      nRead = ReadSocket( sock, buf + nBuf, sizeof buf - nBuf );
      if( nRead < 0 )
         return RC_ERROR;
      if( nRead == 0 )  /* handle last \r (if any) */
         return nBuf == 1 ? WriteBinary( fd, buf, 1, md5ctx ) : RC_OK;
      
      nBuf = WriteAscii( fd, buf, nBuf + nRead, md5ctx );
      if( nBuf == RC_ERROR )
         return nBuf;
      nDtPassed += nRead;
   }
}

/*
   Auxiliary function for Stor, Stou and Appe.
   Stores the data sent by the remote user under the name szPath.
   flags contains bits for open's flags  parameter that are different
   for Stor, Stou and Appe.
*/
int DoStore( int *hDataSocket, const char* szPath, int fd, int flags )
{
   struct stat st;

   if( ReadOnlyBehavior() )
      return SendReply( "550 No permissions\r\n" );

   if( nRestartAt && !strcmp( GetRepresentationType(), "Ascii" ) )
      return SendReply( "425 Restart is prohibited in ASCII mode\r\n" ); 

   if( fd == INVALID_FILE_DESCRIPTOR )
		fd = OpenFile
   	(
      	szPath, O_RDWR | O_CREAT | flags,
      	S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH, /* subject to masking by umask */
      	&st
   	);
   if( fd != INVALID_FILE_DESCRIPTOR )
   {
      if( OpenDataConnection( hDataSocket ) == RC_OK )
      {
         struct timeval start;
         struct MD5Context md5ctx;
         int result;

         timerclear( &start );
         (void) gettimeofday( &start, NULL );
         LogEvent( LLOG_AUDIT, "Stor: %s\n", szPath );

         if( ComputeChecksums() )
            MD5Init( &md5ctx );
         nDtTotal = 0;
         CancelTransferAbort();
         StartTransfer();
         if( !strcmp( GetRepresentationType(), "Ascii" ) )
            result = RecvAsciiFile( fd, *hDataSocket, &md5ctx );
         else
            result = RecvBinaryFile( fd, *hDataSocket, &md5ctx );
         StopTransfer();
         CancelTransferAbort();

         switch( result )
         {
         case RC_ERROR:
         case RC_TIMEDOUT:
            PError( 451 );
            break;
         case RC_ABORTED:
            AccumReply( "426 Data transfer aborted\r\n" );
            AccumReply( "226 File transferred\r\n" );
            break;
         default:
            LogTransferStat( szPath, &start, "Stor", "completed" );
            AccumReply( "226 FILE: " );
            AccumReplyPathname( szPath, strlen( szPath ) );
            AccumReply( " transferred" );
            if( ComputeChecksums() )
            {
               unsigned int i;
               unsigned char sChecksum[ 16 ];
               
               MD5Final( sChecksum, &md5ctx );
               AccumReply( " Checksum (MD5): " );
               for( i = 0; i < sizeof sChecksum; ++i )
                  AccumReplyF( "%02x", sChecksum[ i ] );
            }
            AccumReply( "\r\n" );
            break;
         }
         CloseDataConnection( hDataSocket );
      }
      close( fd );
   }

   /* reset restart point to be sure it won't affect next transfer */
   nRestartAt = 0;
   return SendReplies();
}
