/*
   dir.c

   Incorporates handlers for directory commands.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: dir.c,v 1.20 2008/04/30 23:14:20 shesmax Exp $ */

#define _GNU_SOURCE
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/param.h>  /* MAXPATHLEN */
#include <sys/mman.h>
#include <sys/time.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>      /* for realpath() on glibc systems */
#include <string.h>
#include <malloc.h>

#include "libra.h"
#include "pi.h"
#include "parser.h"
#include "retcodes.h"
#include "log.h"

/*
* Auxiliary functions related to directory and file services
*/

/* 
   Sends the content of the file .message in the current directory 
   through the control connection.
   
   Parameters
   
   code  reply code 
   sPfx  prefix line
*/
int DirInfo( int code, const char* sPfx )
{
   static char sEol[] = "\r\n";
   int fd = open( ".message", O_RDONLY );
   struct stat st;
   char* buf;
   int l;

   /* Send prefix line in any case */
   AccumReplyF( "%3d%c%s\r\n", code, fd == -1 ? ' ' : '-', sPfx );

   if( fd == -1 )
      return SendReplies();
      
   /* Dump the file */
   fstat( fd, &st );
   buf = mmap( 0, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0 );
   if( buf != (void*) -1 )
   {
      char* begin = buf;
      char* end = buf + st.st_size;

      while( begin < end )
      {
         char* p = memchr( begin, '\n', end - begin );
			char* eol = p;

         if( !eol )
            eol = p = end;
			if( eol > begin && eol[ -1 ] == '\r')
				--eol;
         l = eol - begin;
         if
         ( 
            AccumReply( " " ) != RC_OK || 
            AccumReplyNVT( begin, l ) != RC_OK ||
            AccumReply( sEol ) != RC_OK
         )
            break;
         begin = p + 1;
      }  
      munmap( buf, st.st_size );
   }
   close( fd );
   
   /* Send suffix line */
   return SendReplyF( "%3d \r\n", code );
}        

static void DecodePermissions( char* buf, mode_t prot )
{
   static char sPerm[] = "rwxrwxrwx";
   static int anBits[] = 
   {
      S_IRUSR, S_IWUSR, S_IXUSR, 
      S_IRGRP, S_IWGRP, S_IXGRP,
      S_IROTH, S_IWOTH, S_IXOTH
   };
   unsigned int i;

  for( i = 0; i < sizeof anBits / sizeof anBits[ 0 ]; ++i, ++buf )
      if( prot & anBits[ i ] )
         *buf = sPerm[ i ];
}

/* 
   Constructs the line who prefixes the directory entry's name in
   output from ls -l. All of you know what I'm talking about! 
*/
char* LsLongPrefix( const struct stat* st, char* buf, int nBytes )
{
   char sPerm[ 11 ] = "----------" ;
   
   char sTime[ 13 ];
   time_t tLo, tHi;
   size_t n;

   if( nBytes < 57 )
      return 0;

   if( S_ISDIR( st->st_mode ) )
      sPerm[ 0 ] = 'd';
   else if( S_ISCHR( st->st_mode ) )
      sPerm[ 0 ] = 'c';
   else if( S_ISBLK( st->st_mode ) )
      sPerm[ 0 ] = 'b';
   else if( S_ISFIFO( st->st_mode ) )
     sPerm[ 0 ] = 'p';
   else if( S_ISLNK( st->st_mode ) )
     sPerm[ 0 ] = 'l';
   else if( S_ISSOCK( st->st_mode ) )
     sPerm[ 0 ] = 's';
   DecodePermissions( sPerm + 1, st->st_mode );

   /* approximately 6 monthes in the past */
   tLo = time( &tHi ) - 6L * 30L * 24L * 60L * 60L;
   /* one hour in the future */
   tHi += 60L * 60L;
   if( st->st_mtime < tLo || st->st_mtime > tHi )
      /* Mon DD  YYYY */
      n = strftime
      ( 
         sTime, sizeof sTime, "%b %e  %Y", 
         localtime( &st->st_mtime ) 
      );
   else
      /* Mon DD HH:MM */
      n = strftime
      ( 
         sTime, sizeof sTime, "%b %e %H:%M", 
         localtime( &st->st_mtime )
      );
   if( n == 0 || n == sizeof sTime )
      sTime[ 0 ] = '\0';

#if __WORDSIZE == 64 
#define NLINK_FMT "lu"
#else
#define NLINK_FMT "u"
#endif   
   snprintf
   ( 
      buf, nBytes, "%s %4" NLINK_FMT " %-8d %-6d %10ld %12.12s ",
      sPerm, st->st_nlink, 0, 0, st->st_size, sTime   
   );
   return buf;
}

/*
* FTP command handlers
*/

/* Changes the current working directory. */
int Cwd( char* args )
{
   char* szPath;

   if( GetPath( &args, &szPath, GP_OBLIGATORY ) )
      if( chdir( szPath ) == -1 )
         PError( 550 );
      else
         return DirInfo( 250, "Current directory changed" );
   return SendReplies();
}

/* 
   Changes the current working directory to the parent directory. 

   Note: RFC 959 defines 200 as the valid return code for CDUP.
   But this seems not to be logical because the success code for
   CWD is 250. Berkeley's FTP daemon returns 250. So am I.
*/
int CdUp( char* args )
{
   if( ExtraParameters( args ) )
      return SendReplies();
   return Cwd( ".." );
}

/* Removes the specified directory. */
int Rmd( char* args )
{
   char* szPath;

   if( GetPath( &args, &szPath, GP_OBLIGATORY ) )
      if( ReadOnlyBehavior() )
         AccumReply( "550 No permissions\r\n" );
      else if( rmdir( szPath ) == -1 )
         PError( 550 );
      else
      {
         LogEvent( LLOG_AUDIT, "Rmd: %s\n", szPath );
         AccumReply( "250 The directory was successfully deleted\r\n" );
      }
   return SendReplies();
}

/* Creates the specified directory. */
int Mkd( char* args )
{
   char* szPath;

   if( GetPath( &args, &szPath, GP_OBLIGATORY ) )
      if( ReadOnlyBehavior() )
         AccumReply( "550 No permissions\r\n" );
      else if( mkdir( szPath, S_IRWXU | S_IRWXG | S_IROTH | S_IWOTH ) == -1 )
         PError( 550 );
      else
      {
         char buf[ MAXPATHLEN ];

         if( !realpath( szPath, buf ) )
         {
            buf[ 0 ] = '\0';
            strncat( buf, szPath, sizeof buf - 1 );
         }
         LogEvent( LLOG_AUDIT, "Mkd: %s\n", buf );
         AccumReply( "257 \"" );
         AccumReplyNVT( buf, strlen( buf ) );
         AccumReply( "\" directory created\r\n" );
      }
   return SendReplies();
}

/* Returns the reply with the name of the current working directory. */
int Pwd( char* args )
{
   char *wd;

   if( !ExtraParameters( args ) )
   {
      if( (wd = getcwd( 0, 0 )) == 0 )
         PError( 550 );
      else
      {
         AccumReply( "257 \"" );
         AccumReplyPathname( wd, strlen( wd ) );
         AccumReply( "\"\r\n" );
         free( wd );
      }
   }
   return SendReplies();
}
