/*
   dtp.c

   Implementation of Data Transfer Process

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: dtp.c,v 1.31 2002/12/14 23:16:36 shesmax Exp $ */

#include <sys/types.h>

#include <unistd.h>     /* for close */
#include <fcntl.h>      /* for O_ macros */
#include <stdio.h>      /* for sscanf */
#include <stdlib.h>     /* for strtoul */
#include <ctype.h>      /* for toupper */
#include <string.h>     /* for strlen */

#include <sys/socket.h> /* socket related stuff */
#include <netinet/ip.h>
#include <arpa/inet.h>  /* for inet_ntoa */

#include "libra.h"      /* for UNUSED_PARAM */
#include "pi.h"         /* for addrPeerCtrl */
#include "parser.h"     /* for GetPath */
#include "retcodes.h"
#include "dtp.h"
#include "dataopen.h"
#include "chkport.h"
#include "sockets.h"    /* for INVALID_SOCKET */
#include "xbuf.h"

/* Socket related */
int hDataSocket = INVALID_SOCKET;   /* data connection */

/* Connection parameters */
static char cTransferMode;
static char cType;
static char cStructure;
static char cFormat;
static char nByteSize;

/* Maximum mapping length; should be int because of socket interface parameter types. */
int nMaxMap;               
/* no more than MAX_INT/2, see SendAscii(); 128 KB now */
int nAsciiBufSize = 128 * 1024; 

nb_t nDtPassed;            /* current tranfer position */
nb_t nDtTotal;             /* total file size for RETR */
nb_t nRestartAt;           /* restart position */

static int bDataTransferInProgress = 0;
static int bAbortTransfer = 0;

static int bChecksums = RC_FALSE;

/*
* DTP housekeeping
*/

/*
   Initializes the state of the connection.  All parameters retrieve default
   values.  It is assumed that PI is already initialized.
*/
int InitDTP()
{
   bPassive = 0;
   cType = TYPE_ASCII;
   cFormat = FORMAT_NONPRINT;
   cStructure = STRUCTURE_FILE;
   cTransferMode  = TRANSFER_MODE_STREAM;

   nMaxMap = getpagesize();
   nMaxMap = (10L * 1024L * 1024L / nMaxMap) * nMaxMap;  /* approx. 10 MB */

   nRestartAt = 0;
   bDataTransferInProgress = 0;

   /* Init addresses */
   if( (addrSrvrData = SockAddrCreate()) == 0 )
      return RC_FALSE;
   if( (addrPeerData = SockAddrCreate()) == 0 )
      return RC_FALSE;
   SockAddrAssign( addrPeerData, addrPeerCtrl );
   SockAddrSetPort( addrPeerData, DEFAULT_DATA_PORT );
   SrvrDataConn( addrSrvrCtrl->sa_family );
   return RC_TRUE;
}

/* Frees resources allocated for the connection state. */
void ShutdownDTP()
{
   if( hDataSocket != INVALID_SOCKET )
      close( hDataSocket );
   SockAddrDestroy( addrPeerData );
   SockAddrDestroy( addrSrvrData );
}

/*
* FTP command handlers
*/

/* Sets the data representation types.  Only ascii, image and local 8 are supported.  */
int Type( char* args )
{
   char buf[ 16 ];

   if( !*args )
      return SendReply( "501 Need type specifier\r\n" );

   switch( toupper( *args ) )
   {
   case TYPE_ASCII:
      switch( args[ 1 ] )
      {
      case ' ':
         if
         (
            (tolower( args[ 2 ] ) != 'n' && tolower( args[ 2 ] ) != 't') ||
            args[ 3 ]
         )
            goto __error__;
         break;
      case '\0':
         break;
      default:
         goto __error__;
      }
      cType = TYPE_ASCII;
      cFormat = FORMAT_NONPRINT;
      break;
   case TYPE_IMAGE:
      if( args[ 1 ] )
         goto __error__;
      cType = TYPE_IMAGE;
      cFormat = ' '; /* for display purposes only */
      break;
   case TYPE_LOCAL:
      if( args[ 1 ] != ' ' || args[ 2 ] != '8' || args[ 3 ] )
         goto __error__;
      cType = TYPE_LOCAL;
      cFormat = '8'; /* for display purposes only */
      nByteSize = 8;
      break;
   case TYPE_EBCDIC:
      goto __error__;
   }
   snprintf( buf, sizeof buf, "200 Type: %c %c\r\n", cType, cFormat );
   return SendReply( buf );

__error__:
   return SendReply( "504 Supported are: TYPE A [N|T], I and L 8\r\n" );
}

/* Specifies the file structure */
int Stru( char* args )
{
   if( args[ 1 ] )
      return SendReply( "501 Syntax error in parameter\r\n" );
   if( toupper( *args ) != STRUCTURE_FILE )
      return SendReply( "504 Only STRU F is supported\r\n" );
   cStructure = STRUCTURE_FILE;
   return SendReply( "200 Structure is File\r\n" );
}

/* Specifies the transfer mode */
int Mode( char* args )
{
   if( args[ 1 ] )
      return SendReply( "501 Syntax error in parameter\r\n" );
   if( toupper( *args ) != TRANSFER_MODE_STREAM )
      return SendReply( "504 Only MODE S is supported\r\n" );
   cTransferMode = TRANSFER_MODE_STREAM;
   return SendReply( "200 Current transfer mode is Stream" );
}

/* Sends the given file to the remote host */
int Retr( char* args )
{
   char* szPath;

   if( !GetPath( &args, &szPath, GP_OBLIGATORY ) )
     return SendReplies();
   return DoRetrieve( &hDataSocket, szPath );
}

/*
   Accept the data sent by the user DTP and store it as the file with
   the given name. If such file exists, overwrite it.
*/
int Stor( char* args )
{
   char* szPath;

   if( !GetPath( &args, &szPath, GP_OBLIGATORY ) )
     return SendReplies();
   return DoStore( &hDataSocket, szPath, INVALID_FILE_DESCRIPTOR, nRestartAt ? 0 : O_TRUNC );
}

/*
   Stores the data sent by the remote user under the name unique in the
   current directory.
*/
int Stou( char* args UNUSED_PARAM )
{
   /* RFC 959 specifies that STOU has no parameters, but the
      Berkeley's FTP client sends the name as a parameter to the
      command. In accordance to the rule "Be liberal in what you
      accept ... " I decided to discard that parameter and work
      as if there was no extra input after the command. */
   int result;
   char name[] = "./libraXXXXXX";
   int fd;

   if( (fd = mkstemp( name )) != -1 )
		/* do not allow to overwrite an existing file */
      result = DoStore( &hDataSocket, name, fd, O_EXCL );       
	else {
      PError( 450 );
      result = RC_ERROR;
   }
   return result;
}

/* Appends the data sent by the remote user at the end of the given file. */
int Appe( char* args )
{
   char* szPath;

   if( !GetPath( &args, &szPath, GP_OBLIGATORY ) )
      return SendReplies();
   return DoStore( &hDataSocket, szPath, INVALID_FILE_DESCRIPTOR, O_APPEND );
}

/* Reserve storage at the server side. */
int Allo( char* args UNUSED_PARAM )
{
   return SendReply( "202 Command not implemented, superfluous at this site.\r\n" );
}

int Rest( char* args )
{
   char* endp;
   nb_t value;
   unsigned long arg;

   if( !*args )
      return SendReply( "501 Need a stream offset\r\n" );

   arg = strtoul( args, &endp, 10 );
   value = arg;
   if( *endp || (off_t)value < 0 || arg != value ) /* type independent check */
      return SendReply( "501 Illegal number\r\n" );
   nRestartAt = value;
   return SendReply( "350 Ok, send file transfer request\r\n" );
}

/*
   Query and control DTP
*/

char* GetRepresentationType()
{
   switch( cType )
   {
   case TYPE_ASCII:
      return "Ascii";
   case TYPE_IMAGE:
      return "Image";
   default:
      return "Local 8";
   }
}

char* GetFileStructure()
{
   return "File";
}

char* GetTransferMode()
{
   return "Stream";
}

int DataTransferInProgress()
{
   return bDataTransferInProgress;
}

void StartTransfer()
{
   bDataTransferInProgress = RC_TRUE;
}

void StopTransfer()
{
   bDataTransferInProgress = RC_FALSE;
}

int AbortTransfer()
{
   return bAbortTransfer;
}

void PostTransferAbort()
{
   bAbortTransfer = RC_TRUE;
}

void CancelTransferAbort()
{
   bAbortTransfer = RC_FALSE;
}

nb_t DataToSend()
{
   return nDtTotal;
}

nb_t DataSent()
{
   return nDtPassed;
}

int ComputeChecksums()
{
   return bChecksums;
}

void SetChecksumMode( int on )
{
   bChecksums = on;
}
