/*
   dtplist.c

   Implementation of LIST and NLST commands.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: dtpls.c,v 1.10 2008/04/30 23:14:20 shesmax Exp $ */

#include <string.h>     /* for strlen */
#include <stdlib.h>

#include "libra.h"      /* for UNUSED_PARAM */
#include "retcodes.h"
#include "pi.h"         /* AccumReply, PError */
#include "dtp.h"        /* all DTP related stuff */
#include "dataopen.h"
#include "parser.h"     /* GetPath, GP_... macros */
#include "listdir.h"    /* Ls, LsCallback */
#include "xbuf.h"       /* XBuf related things */
#include "utf8.h"

/*
   Callback function for LIST command.

   Parameters

   name     File name
   st       File attributes
   param    Pointer to DCLP structure

   Returns

   RC_OK          No errors
   RC_ERROR       OS error occured.
   RC_TIMEDOUT    Waiting for new data timed out.
   RC_ABORTED     Data transfer aborted by the user.
*/
static int LongListEntry( const char* name, const struct stat* st, void* param )
{
   char sPfx[ 57 ];
   int nResult;
   XBuf* xb = ((DCLP*)param)->xb;
   int socket = ((DCLP*)param)->socket;
   const char* fname;
   size_t len;

   LsLongPrefix( st, sPfx, sizeof sPfx );
   nResult = PutSockXBuf( xb, sPfx, sizeof sPfx - 1, FlushSockXBuf, socket );
   if( nResult != RC_OK )
      return nResult;
   fname = strrchr( name, '/' );
   if( !fname )
      fname = name;
   else
      ++fname;
      
   /* If UTF-8 encoded pathnames are required for the data connection then we
    * need to convert pathnames that are coming from filesystem to UTF-8 
    * prior to send them to the User DTP. Such conversion is only necessary 
    * if a pathname contains characters from a local character set. ASCII 
    * characters can be transmitted without any conversion.
    *
    * If UTF-8 encoding of pathnames is not required we send pathnames as is.
    * This behavior violates the requirement of RFC959 to send only ASCII 
    * characters in pathnames. It was intentionally programmed so in order to
    * provide the possibility to transmit files with language-specific names.
    * This is working well of cause  only if a client and the server use  
    * the very same character set. 
    *
    * If filesystem is using UTF-8 encoded pathnames then they are also sent 
    * without any modification to the client program. This was done so because 
    * the server has no way to determine the character set of the client and 
    * thus cannot do any meaningfull conversion of pathnames. One option would
    * be to replace all non-ASCII characters with some ASCII character, but
    * this would give no chance to the client to convert a pathname properly. */
   len = strlen( fname );
   if( utf8_enctype( (const unsigned char*)fname, len ) == ENCTYPE_LOCAL &&
       UTF8PathnamesOverDataConnection() == RC_TRUE )
   {
      unsigned char* utf8;

      if( !(len = local_to_utf8( (const unsigned char*)fname, len, &utf8 )) )
         nResult = RC_ERROR;
      else {
         nResult = PutSockXBuf( xb, (const char*)utf8, len, FlushSockXBuf, socket );
         free( utf8 );
      }
   }
   else 
      nResult = PutSockXBuf( xb, fname, len, FlushSockXBuf, socket );
   return nResult == RC_OK ? PutSockXBuf( xb, "\r\n", 2, FlushSockXBuf, socket ) : nResult;
}

/*
   Callback function for NLST command.

   Parameters

   name     File name
   st       File attributes
   param    Pointer to DCLP structure

   Returns

   RC_OK          No errors
   RC_ERROR       OS error occured.
   RC_TIMEDOUT    Waiting for new data timed out.
   RC_ABORTED     Data transfer aborted by the user.
*/
static int NameListEntry( const char* name, const struct stat* st UNUSED_PARAM, void* param )
{
   XBuf* xb = ((DCLP*)param)->xb;
   int socket = ((DCLP*)param)->socket;
   int nResult;
   size_t len;
      
   /* If UTF-8 encoded pathnames are required for the data connection then we
    * need to convert pathnames that are coming from filesystem to UTF-8 
    * prior to send them to the User DTP. Such conversion is only necessary 
    * if a pathname contains characters from a local character set. ASCII 
    * characters can be transmitted without any conversion.
    *
    * If UTF-8 encoding of pathnames is not required we send pathnames as is.
    * This behavior violates the requirement of RFC959 to send only ASCII 
    * characters in pathnames. It was intentionally programmed so in order to
    * provide the possibility to transmit files with language-specific names.
    * This is working well of cause  only if a client and the server use  
    * the very same character set. 
    *
    * If filesystem is using UTF-8 encoded pathnames then they are also sent 
    * without any modification to the client program. This was done so because 
    * the server has no way to determine the character set of the client and 
    * thus cannot do any meaningfull conversion of pathnames. One option would
    * be to replace all non-ASCII characters with some ASCII character, but
    * this would give no chance to the client to convert a pathname properly. */
   len = strlen( name );
   if( utf8_enctype( (const unsigned char*)name, len ) == ENCTYPE_LOCAL &&
         UTF8PathnamesOverDataConnection() == RC_TRUE )
   {
      unsigned char* utf8;

      if( !(len = local_to_utf8( (const unsigned char*)name, len, &utf8 )) )
         nResult = RC_ERROR;
      else {
         nResult = PutSockXBuf( xb, (const char*)utf8, len, FlushSockXBuf, socket );
         free( utf8 );
      }
   }
   else
      nResult = PutSockXBuf( xb, name, strlen( name ), FlushSockXBuf, socket );
   
   if( nResult != RC_OK )
      return nResult;
   return PutSockXBuf( xb, "\r\n", 2, FlushSockXBuf, socket );
}


/*
   Auxiliary function for List(), Nlst() and Mlsd()

   path        Pathname.
   callback    Callback function
   params      User-supplied parameters
*/
int DtpLs( char* szPath, LsCallback callback, void* params )
{
   XBuf xbuf;

   if( CreateFixedXBuf( &xbuf, nAsciiBufSize ) == XBUF_OK )
   {
      if( OpenDataConnection( &hDataSocket ) == RC_OK )
      {
         int nResult;
         DCLP dlcp;
         
         dlcp.xb = &xbuf;
         dlcp.socket = hDataSocket;
         dlcp.params = params;
         CancelTransferAbort();
         StartTransfer();

         nResult = Ls( szPath, callback, (void*)&dlcp );
         if( nResult == RC_OK && xbuf.nUsed )
            nResult = FlushSockXBuf( &xbuf, hDataSocket );

         StopTransfer();
         CancelTransferAbort();

         switch( nResult )
         {
         case RC_ERROR:
         case RC_TIMEDOUT:
            PError( 451 );
            break;
         case RC_ABORTED:
            AccumReply( "426 Transfer aborted\r\n" );
            /* fall through */
         case RC_OK: 
            AccumReply( "226 Transfer completed\r\n" );
            break;
         }

         CloseDataConnection( &hDataSocket );
      }
   
      DestroyXBuf( &xbuf );
   }
   else
      PError( 451 );

   return SendReplies();
}

/*
   Auxiliary function for LIST. It skips any options the FTP client may pass
   to it. This behavior of the client is not specified in RFC. And it is not
   portable to platforms other then UNIX. 
*/
static char* SkipOptions( char* args )
{
   int bInWord = 0;
   
   while( *args )
   {
      if( *args == ' ' )
         bInWord = 0;
      else if( *args == '-' )
      {
         if( !bInWord && args[ 1 ] == '-' && args[ 2 ] == ' ' )
         {
            args += 3;
            while( *args && *args == ' ' )
               ++args;
            break;
         }
         bInWord = 1;
      }
      else
      {
         if( !bInWord )
            break;
         bInWord = 1;
      }
      ++args;
   }
   return args;
}

int List( char* args )
{
   char* szPath;
   
   args = SkipOptions( args );
   GetPath( &args, &szPath, GP_OPTIONAL );
   return DtpLs( szPath, LongListEntry, 0 );
}

int Nlst( char* args )
{
   char* szPath;
   
   GetPath( &args, &szPath, GP_OPTIONAL );
   return DtpLs( szPath, NameListEntry, 0 );
} 
