/*
   inet6_xtoy.c 
   
   Implementation of inet6_aton().
   Converts a standard string representation of IPv6 address to binary data.
   Implementation of inet6_ntoa().
   Converts binary representation of IPv6 address to a text string.

   Copyright ( C ) 2001   Andrey Savochkin ( saw@saw.sw.com.sg )
                        Maxim Shesterikov ( libraftp@narod.ru )

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: inet6_xtoy.c,v 1.7 2002/05/18 05:50:02 shesmax Exp $ */

#include <netinet/in.h>     /* for struct in6_addr */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define ae_err  1
#define ae_nul  2
#define ae_hex  3
#define ae_dec  4

typedef int addr_elem_t;

addr_elem_t get_addr_elem(  char** as, uint16_t* elem, addr_elem_t last_type  )
{
   char* e;
   addr_elem_t cur_type;
   uint16_t x;

   /* Find the end of current address element */
   for(  e = *as;; ++e  )
      if( *e == '\0' ) 
      {
         cur_type = last_type;
         break;
      } 
      else if ( *e == ':' ) 
      {
         if( last_type == ae_dec )
            return ae_err;                /* X:d.d:X */
         if( *as == e ) 
         {                                 /* X::X" */
            *as = e + 1;
            return ae_nul;
         }
         if( !e[ 1 ] )
            return ae_err;                /* X:X: */
         cur_type = ae_hex;
         break;
      } 
      else if( *e == '.' ) 
      {
         if( *as == e )                    /* .10, X:.10, 10..31 */
            return ae_err;
         if( !e[ 1 ] )
            return ae_err;                /* 10.31.1. */
         cur_type = ae_dec;
         break;
      }

   x = 0;
   if( cur_type == ae_hex ) 
   {
      if( e - *as > 4 )
         return ae_err;    /* Address element is too long - hhhh is max */
      while( *as < e ) 
      {
         if( !isascii( **as ) || !isxdigit( **as ) )
            return ae_err; /* Wrong character */
         x <<= 4;
         if( isdigit( **as ) )
            x += **as - '0';
         else
            x += toupper( **as ) - 'A' + 10;
         ++*as;
      }
   } 
   else 
   {
      if( e - *as > 3 )
         return ae_err;    /* Address element is too long - ddd is max */
      while( *as < e ) 
      {
         if( !isascii( **as ) || !isdigit( **as ) )
            return ae_err; /* Wrong character */
         x = x * 10 + ( **as - '0' );
         ++*as;
      }
      if( x > 255 )
         return ae_err;    /* X:X:X:X:X:X:193.212.555.250 */
   }
   if( *e != '\0' )
      *as = e + 1;
   *elem = x;
   return cur_type;
}

#define IP6_ADDR_WORDS  8
#define IP4_ADDR_WORDS  2
#define IP4_ADDR_BYTES  4
int inet6_aton( char* as, struct in6_addr* addr )
{
   uint16_t elems[ IP6_ADDR_WORDS ];
   uint16_t elem;
   int nelems = 0;
   addr_elem_t cur = ae_err;

   int nnuls = 0;
   int nleft = 0;

   int ndec = 0;
   int nhex;

   int src, dst;

   if( !*as )
      return -1;     /* No address */
   if( *as == ':' ) 
   {
      if( as[1] != ':' )                      /* :X:X:X */
         return 0;
      ++as;                                  /* ::X */
   }

   /* Parse address string as and put address elements into elems array */
   while( *as ) 
   {
      switch( cur = get_addr_elem( &as, &elem, cur ) ) 
      {
      case ae_err:
         return 0;
      case ae_nul:
         if( nnuls )
            return 0;                    /* X::X::X */
         ++nnuls;
         nleft = nelems;
         cur = ae_hex;
         break;
      case ae_hex:
         if( nelems == IP6_ADDR_WORDS )
            return 0;
         elems[ nelems++ ] = elem;
         break;

      case ae_dec:
         if( ndec == IP4_ADDR_BYTES )
            return 0;           /* ::X:d.d.d.d.d */
         switch( ++ndec ) 
         {
         case 1:
            if( nelems > IP6_ADDR_WORDS - IP4_ADDR_WORDS )
               return 0;
            elems[ 6 ] = elems[ 7 ] = 0;
            elems[ 6 ] |= elem << 8;
            break;
         case 2:
            elems[ 6 ] |= elem;
            break;
         case 3:
            elems[ 7 ] |= elem << 8;
            break;
         case 4:
            elems[ 7 ] |= elem;
            break;
         }
         break;
      }
   }

   /* Check number of elements */
   if( ndec ) 
   {
      if( ndec != 4 )
         return 0;                 /* X:X:X:X:X:X:d.d.d */
      nhex = IP6_ADDR_WORDS - IP4_ADDR_WORDS;
   } 
   else
      nhex = IP6_ADDR_WORDS;
   if( !nnuls ) 
   {
      if( nelems != nhex )
         return -1;
      nleft = nelems;
   } 
   else if( nelems == nhex )
      return 0;                    /* no place for zero bytes */

   /* Copy IPv6 elements to address memory block. Align left and right parts
   of address before and after :: properly */
   for( src = 0; src < (int)(sizeof elems / sizeof elems[ 0 ]); ++src )
      elems[ src ] = htons( elems[ src ] );
   memset( addr->s6_addr16, 0, sizeof elems );
   dst = 0;
   for( src = 0; src < nleft; ++src, ++dst )
      addr->s6_addr16[ dst ] = elems[ src ];
   for( dst = nhex - nelems + nleft; src < nelems; ++src, ++dst )
      addr->s6_addr16[ dst ] = elems[ src ];
   /* Copy decimal elements ( if any ) */
   if( ndec ) 
   {
      addr->s6_addr16[ 6 ] = elems[ 6 ];
      addr->s6_addr16[ 7 ] = elems[ 7 ];
   }
   return 1;
}

int inet6_ntoa( struct in6_addr *addr, char *buf, size_t sz  )
{
   static char ipv4_compat[ 12 ] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
   static char ipv4_mapped[ 12 ] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF};
   int nshorts = 0;
   int i;
   char smallbuf[ 6 ];
   size_t sblen;

   if( memcmp( addr->s6_addr, ipv4_compat, sizeof ipv4_compat ) == 0 )
      return snprintf
      ( 
         buf, sz, "%d.%d.%d.%d", 
         addr->s6_addr[ 12 ], addr->s6_addr[ 13 ],
         addr->s6_addr[ 14 ], addr->s6_addr[ 15 ]
      );
   if( memcmp( addr->s6_addr, ipv4_mapped, sizeof ipv4_mapped ) == 0 )
      return snprintf
      ( 
         buf, sz, "::FFFF:%d.%d.%d.%d", 
         addr->s6_addr[ 12 ], addr->s6_addr[ 13 ],
         addr->s6_addr[ 14 ], addr->s6_addr[ 15 ]
      );

   buf[ 0 ] = '\0';
   for( i = 0; i < IP6_ADDR_WORDS; ++i )
   {
      smallbuf[ 0 ] = '\0';
      sblen = 0;
      if( nshorts == 0 && addr->s6_addr16[ i ] == 0 )
      {
         int last;
         
         for( last = i + 1; last < IP6_ADDR_WORDS - 1; ++last )
            if( addr->s6_addr16[ last ] != 0 )
               break;
         if( last - i > 1 )
         {
            if( buf[ 0 ] == '\0' )
            {
               smallbuf[ 0 ] = ':';
               sblen = 1;
            }
            smallbuf[ sblen++ ] = ':';
            smallbuf[ sblen ] = '\0';
            i = last;
            nshorts = 1;
         }
      }

      sblen += sprintf( smallbuf+sblen, "%X", ntohs( addr->s6_addr16[ i ] ) );
      if( i < IP6_ADDR_WORDS - 1 )
      {
         smallbuf[ sblen++ ] = ':';
         smallbuf[ sblen ] = '\0';
      }
      if( sblen >= sz )
         return -1;
      strcat( buf, smallbuf );
      sz -= sblen;
   }
   return 0;
}
