/*
   log.c

   Implementation of log facility. Writes log messages to stderr.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: log.c,v 1.15 2002/05/18 05:50:02 shesmax Exp $ */

#include <unistd.h>
#include <errno.h>
#include <sys/utsname.h>
#include <time.h>

#include <stdlib.h>

#include "log.h"
#include "retcodes.h"
#include "xbuf.h"

static XBuf emsg;       /* Expandable buffer holds the message */
static int nPfxLen = 0; /* The length of the log message's prefix */
static int bInited = RC_FALSE;
static int events = _LLOG_ERROR | _LLOG_WARN;

#define INI_SIZE  160

int OpenLog( const char* sPgmName )
{
   struct utsname buf;
   char* nodename;
   
   if( bInited )
      return RC_TRUE;
   bInited = RC_FALSE;
   
   if( CreateExpandableXBuf( &emsg, INI_SIZE ) == XBUF_OVERFLOW )
      return RC_FALSE;
   bInited = RC_TRUE;
   
   /* 
      Get the name of node either from environment (for virtual
      hosts) or system wide.
   */      
   nodename = getenv( "TCPLOCALHOST" ); /* tcp server specific */
   if( !nodename )
      if( uname( &buf ) != -1 )
         nodename = buf.nodename;
   
   PrintfXBuf( &emsg, "%s|%s[%d]|", nodename, sPgmName, getpid() );    
   nPfxLen = emsg.nUsed;
   
   events = _LLOG_ERROR | _LLOG_WARN;
   
   return RC_TRUE;
}

int CloseLog()
{
   if( bInited )
   {
      DestroyXBuf( &emsg );
      bInited = RC_FALSE;
   }
   return RC_TRUE;
}

int ConfigureLog( const char* params )
{
   const char* p;

   if( !bInited )
      return RC_ERROR;

   for( p = params; *p; ++p )
      switch( *p )
      {
      case 'i':
         events |= _LLOG_CMD;
         break;
      case 'a':
         events |= _LLOG_AUDIT;
         break;
      case 'u':
         events |= _LLOG_CONNECT;
         break;
      case 'o':
         events |= _LLOG_REPLY;
         break;
      default:
         return p - params;
      }
      
   return RC_OK;
}

int LogEvent( int event, const char* fmt, ... )
{
   va_list args;
   time_t t;
   char sTime[ 80 ];
   size_t nTimeChars;
   size_t l;
   ssize_t r;
   int result;

   if( !(events & event) )
      return RC_OK;
   switch( LOG_EVENT( event ) )
   {
   case _LLOG_CMD:
      event = 'i';
      break;
   case _LLOG_AUDIT:
      event = 'a';
      break;
   case _LLOG_CONNECT:
      event = 'u';
      break;
   case _LLOG_REPLY:
      event = 'o';
      break;
   case _LLOG_ERROR:
      event = 'e';
      break;
   case _LLOG_WARN:
      event = 'w';
      break;
   default:
      event = '?';
      break;
   }
   
   va_start( args, fmt );

   emsg.nUsed = nPfxLen;
   t = time( 0 );
   nTimeChars = strftime( sTime, sizeof sTime, "%c", localtime( &t ) );
   if( nTimeChars <= 0 || nTimeChars >= sizeof sTime )
      sTime[ 0 ] = '\0';

   if
   ( 
      PrintfXBuf( &emsg, "%s|%c|", sTime, (char)event ) == XBUF_OVERFLOW ||
      VPrintfXBuf( &emsg, fmt, args ) == XBUF_OVERFLOW
   )
      result = RC_ERROR;
   else
   {
      result = RC_OK;
      for( l = 0; l < emsg.nUsed; )
      {
         r = write( STDERR_FILENO, emsg.data + l, emsg.nUsed - l );
         if( r > 0 )
            l += r;
         else if( errno != EAGAIN )
         {
            result = RC_ERROR;
            break;          
         }
      }
   }

   va_end( args );   
   if( result == RC_OK )
      result = event & _LLOG_FATAL ? RC_ERROR : RC_OK;
   return result;
}
