/* 
	mlsx.c

   Implemetation of MLST and MLSD commands.

   Copyright (C) 2000   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: mlsx.c,v 1.14 2008/04/28 05:37:54 shesmax Exp $ */

#include <unistd.h>     
#include <sys/stat.h>
#include <string.h>     
#include <stdio.h>
#include <stdlib.h>

#include "mlsx.h"
#include "libra.h"      /* for UNUSED_PARAM */
#include "pi.h"         
#include "parser.h"     /* for GetPath */
#include "retcodes.h"
#include "dtp.h"
#include "dataopen.h"
#include "listdir.h"
#include "xbuf.h"
#include "utf8.h"

/* 
   Facts
*/
#define FACT_CHARSET       1
#define FACT_CREATE        2
#define FACT_LANG          4
#define FACT_MEDIA_TYPE    8
#define FACT_MODIFY        16
#define FACT_PERM          32
#define FACT_SIZE          64
#define FACT_TYPE          128
#define FACT_UNIQUE        256

unsigned long nActiveFacts = 
   FACT_MODIFY | FACT_PERM | FACT_SIZE | FACT_TYPE | FACT_UNIQUE;  

/* Note that the array should be sorted */
static char* facts[] = 
{
   "charset",
   "create", 
   "lang",
   "media-type", 
   "modify", 
   "perm", 
   "size", 
   "type", 
   "unique", 
};

/* For brevity */
typedef const struct stat* PCSTAT;
typedef const char* PCCHAR;
typedef int (*FactFunc)( XBuf*, PCSTAT, PCCHAR path, PCSTAT );    

static int FactType( XBuf*, PCSTAT, PCCHAR, PCSTAT );
static int FactSize( XBuf*, PCSTAT, PCCHAR, PCSTAT );
static int FactUnique( XBuf*, PCSTAT, PCCHAR, PCSTAT );
static int FactModify( XBuf*, PCSTAT, PCCHAR, PCSTAT );
static int FactPerm( XBuf*, PCSTAT, PCCHAR, PCSTAT );

FactFunc pfFacts[] =
{
   0,             /* charset */
   0,             /* create */ 
   0,             /* lang */
   0,             /* media-type */
   FactModify,    /* modify */
   FactPerm,      /* perm */
   FactSize,      /* size */ 
   FactType,      /* type */ 
   FactUnique     /* unique */
};   

#define DIM( a ) (sizeof a / sizeof a[ 0 ])

/* 
   bAllFacts is true when called from FEAT when we need to print the names
   of all supported facts by marking active with asterics. Otherwise it is
   called by OPTS. In this case we need to print out active facts only.
*/   
void MlsxListFacts( int bAllFacts )
{
   unsigned int i;
   unsigned long facts_mask = bAllFacts ? (unsigned long)~0 : nActiveFacts;
   
   AccumReply( " " );
   for( i = 0; i < DIM( facts ); ++i )
      if( pfFacts[ i ] && (facts_mask & (1 << i)) )
      {
         AccumReplyBuf( facts[ i ], strlen( facts[ i ] ) );
         if( facts_mask & (1 << i) )
            AccumReply( "*" );
         AccumReply( ";" );
      }
}

static int FactCmp( const void* key, const void* memb )
{
   return strcasecmp( (const char*)key, *(const char**)memb );
}

/* Returns the list of facts supported by the server */
#include "log.h"
int MlsxOpts( char* desired_facts )
{
   int nNewFacts = 0;
   char* fact;
   char *end = desired_facts + strlen( desired_facts );

   if( *desired_facts == ' ' )
      ++desired_facts;
   for( fact = desired_facts; fact < end; ++desired_facts )
      if( !*desired_facts || *desired_facts == ';' )
      {
         char** elem;
         int index;

         *desired_facts = '\0';
         elem = (char**)bsearch 
         (          
            fact, facts, DIM( facts ), sizeof facts[ 0 ], FactCmp
         );
         if( !elem )
            return SendReply( "501 Invalid MLST fact\r\n" );
         index = elem - facts;
         if( pfFacts[ index ] )
            nNewFacts |= 1 << index;
         fact  = desired_facts + 1;
      }
 
   nActiveFacts = nNewFacts;
   
   AccumReply( "200 MLST OPTS" );
   if( nActiveFacts )
      MlsxListFacts( MLF_ACTIVE_FACTS );
   return SendReply( "\r\n" );
}

/*
   Building fact texts
*/

static int FactType
( 
   XBuf* xb, 
   PCSTAT st, 
   PCCHAR path, 
   PCSTAT parent UNUSED_PARAM 
)
{
   char* type;
   mode_t m = st->st_mode;
      
   if( path[ 0 ] == '.' && path[ 1 ] == '\0' )
      type = "cdir";
   else if( !strcmp( path, ".." ) )
      type = "pdir";
   else if( S_ISDIR( m ) )
      type = "dir";
   else if( S_ISREG( m ) )
      type = "file";
   else if( S_ISCHR( m ) )
      type = "UNIX.char";
   else if( S_ISBLK( m ) )
      type = "UNIX.block";
   else if( S_ISFIFO( m ) )
      type = "UNIX.fifo";
   else if( S_ISSOCK( m ) )
      type = "UNIX.socket";
   else
      return RC_OK;
   return PrintfXBuf( xb, "type=%s;", type ) > XBUF_OK ? RC_OK : RC_ERROR;
}   

static int FactSize
( 
   XBuf* xb, 
   PCSTAT st, 
   PCCHAR path UNUSED_PARAM, 
   PCSTAT parent UNUSED_PARAM
)
{
   if( S_ISREG( st->st_mode ) )
      return PrintfXBuf( xb, "size=%d;", st->st_size ) > XBUF_OK ? 
         RC_OK : RC_ERROR;
   return RC_OK;
}   

static int FactUnique
( 
   XBuf* xb, 
   PCSTAT st, 
   PCCHAR path UNUSED_PARAM, 
   PCSTAT parent UNUSED_PARAM
)
{
   return PrintfXBuf( xb, "unique=%d;", st->st_ino ) > XBUF_OK ? 
      RC_OK : RC_ERROR;
}

static int FactModify
( 
   XBuf* xb, 
   PCSTAT st, 
   PCCHAR path UNUSED_PARAM, 
   PCSTAT parent UNUSED_PARAM 
)
{
   struct tm* tmP = gmtime( &st->st_mtime );

   return PrintfXBuf
   ( 
      xb, "modify=%04d%02d%02d%02d%02d%02d;",
      tmP->tm_year + 1900, tmP->tm_mon + 1, tmP->tm_mday,
      tmP->tm_hour, tmP->tm_min, tmP->tm_sec
   ) > XBUF_OK ? RC_OK : RC_ERROR;
}

/* Relationship to the object */
#define RL_OWNER  0
#define RL_GROUP  1
#define RL_OTHER  2

/* Determines the relationship to the directory object */
static int WhoAmI( const struct stat* st )
{
   if( st->st_uid == geteuid() )
      return RL_OWNER;
   if( st->st_gid == getgid() )
      return RL_GROUP;
   return RL_OTHER;
}

/* Facilitates to determine whether the process has the right to do smth. */
static int nRdPermMask[ 3 ] = {S_IRUSR, S_IRGRP, S_IROTH};
static int nWrPermMask[ 3 ] = {S_IWUSR, S_IWGRP, S_IWOTH};
static int nExPermMask[ 3 ] = {S_IXUSR, S_IXGRP, S_IXOTH};

/* 
   Returns true if current user can either read, write or 
   execute the object represented by st.
*/   
static int CanDo( const struct stat* st, int* what, int whoami )
{
   mode_t perm = st->st_mode;

   while( whoami <= RL_OTHER )
      if( perm & what[ whoami++ ] )
         return RC_TRUE;
   return RC_FALSE;
}

#define CAN_READ( st ) CanDo( (st), nRdPermMask, whoami )
#define CAN_WRITE( st ) CanDo( (st), nWrPermMask, whoami )
#define CAN_LIST( st ) CanDo( (st), nExPermMask, whoami )

static int FactPerm
( 
   XBuf* xb, 
   PCSTAT st, 
   PCCHAR path UNUSED_PARAM, 
   PCSTAT parent 
)
{
   int whoami = WhoAmI( st );
   int r = CAN_READ( st );
   int w = CAN_WRITE( st );
   int x = CAN_LIST( st );
#define MAX_NUMBER_OF_PERMISSIONS   10
   char perm[ MAX_NUMBER_OF_PERMISSIONS + 1 ];
#undef MAX_NUMBER_OF_PERMISSIONS
   char* p = perm;
  
   if( S_ISDIR( st->st_mode ) )
   {
      if( w )
      {
         *p++ = 'c';             /* can create objects there */
         *p++ = 'm';             /* can rename objects there */
         *p++ = 'p';             /* is able to delete objects there */
      }
      if( x )
      {
         memcpy( p, "el", 2 );   /* can enter and list content */
         p += 2;
      }
   }
   else if( S_ISREG( st->st_mode ) ) /* regular file */
   {
      if( r )
         *p++ = 'r';             /* permission to read */
      if( w )
      {
         memcpy( p, "aw", 2 );   /* permission to append and write */
         p += 2;
      }
   }

   /* Determine the right in parent directory */
   whoami = WhoAmI( parent );
   if( CAN_WRITE( parent ) )
   {
      *p++ = 'f';                /* permission to rename the object */
      *p++ = 'd';                /* permission to delete the object */
   }
   *p = '\0';

   return PrintfXBuf( xb, "perm=%s;", perm ) > XBUF_OK ? RC_OK : RC_ERROR;
}

/* Concatenates all active facts together */
static int Facts( XBuf* xb, PCSTAT obj, PCCHAR path, PCSTAT parent )    
{
   unsigned int i;
   int rc = RC_OK;

   for( i = 0; i < DIM( facts ); ++i )
      if( nActiveFacts & (1 << i) )
      {
         rc = (pfFacts[ i ])( xb, obj, path, parent );
         if( rc != RC_OK )
            break;
      }
   return rc;
}

/*
    Implementation of MLSD.  
*/

/*
   Callback function for MLSD command.

   Parameters

   name     File name
   st       File attributes
   param    Pointer to struct MlsdEntryParams

   Returns

   RC_OK          No errors
   RC_ERROR       OS error occured.
   RC_TIMEDOUT    Waiting for new data timed out.
   RC_ABORTED     Data transfer aborted by the user.
*/
static int MlsdEntry( const char* name, const struct stat* obj, void* param )
{
   XBuf* xb = ((DCLP*)param)->xb;
   int sock = ((DCLP*)param)->socket;
   struct stat* parent = ((DCLP*)param)->params;
   XBuf buf;
   int rc;
   const char* fname;
   size_t len;
   
   if( CreateExpandableXBuf( &buf, 256 ) != XBUF_OK )
      return RC_ERROR;
   do
   {
      if( (rc = Facts( &buf, obj, name, parent )) != RC_OK )
         break;
      rc = PutSockXBuf( xb, buf.data, buf.nUsed, FlushSockXBuf, sock );
      if( rc != RC_OK )
         break;
      rc = PutSockXBuf( xb, " ", 1, FlushSockXBuf, sock );
      if( rc != RC_OK )
         break;

      fname = strrchr( name, '/' );
      if( fname )
         ++fname;
      else
         fname = name;
      len = strlen(fname);
      if( utf8_enctype( (const unsigned char*)fname, len ) != ENCTYPE_ASCII )
      {
         unsigned char* utf8;

         if( !(len = local_to_utf8( (const unsigned char*)fname, len, &utf8 )) )
         {
            rc = RC_ERROR;
            break;
         }
         rc = PutSockXBuf( xb, (const char*)utf8, len, FlushSockXBuf, sock );
         free( utf8 );
      }
      /* No conversion is necessary */
      else 
         rc = PutSockXBuf( xb, fname, len, FlushSockXBuf, sock );
      if( rc != RC_OK )
			break;
      rc = PutSockXBuf( xb, "\r\n", 2, FlushSockXBuf, sock );
   }
   while( 0 );
   DestroyXBuf( &buf );
   return rc;
}

/* 
   Returns the statistics about the directory which is a parent of
   the given path. 
*/
static int StatParent( char* path, struct stat* st )
{
   int rc;
   char* parent = path;
   char* pC = strrchr( path, '/' );
   char c = 0;  /* do not remove the initialization - it satisfies gcc! */

   if( !pC )               /* directory name */
      parent = "..";
   else if( pC == path )   /* root directory */
      pC = 0;
   else                    /* path name */
   {
      c = *pC;
      *pC = '\0';
   }
   rc = stat( parent, st );
   if( pC )
      *pC = c;
   return rc;
}


int Mlsd( char* args )
{
   char* szPath;
   struct stat obj;
   
   if( !GetPath( &args, &szPath, GP_OPTIONAL )  )
      return SendReplies();
   if( !*szPath )
      szPath = ".";
   if( stat( szPath, &obj ) == -1 )
      PError( 550 );  
   else if( !S_ISDIR( obj.st_mode ) )
      AccumReply( "550 Argument must be a directory\r\n" );
   else
   {
      struct stat parent;

      if( StatParent( szPath, &parent ) == -1 )
         PError( 550 );
      else
         return DtpLs( szPath, MlsdEntry, (void*)&parent );
   }
   return SendReplies();
}

/*
    Implementation of MLST.  
*/
int Mlst( char* args )
{
   static char szNoMem[] = "550 The system has scarce memory resources\r\n";
   char* szPath;
   struct stat obj;
   XBuf buf;
   
   if( !GetPath( &args, &szPath, GP_OPTIONAL )  )
      return SendReplies();
   if( !*szPath )
      szPath = ".";
   
   if( stat( szPath, &obj ) == -1 )
      PError( 550 );   
   else if( CreateExpandableXBuf( &buf, 256 ) == XBUF_OK )
   {
      struct stat parent;

      if( StatParent( szPath, &parent ) == -1 )
         PError( 550 );
      else if( Facts( &buf, &obj, szPath, &parent ) == RC_OK )
      {
         int pathlen = strlen( szPath );
      
         AccumReply( "210 " );
         AccumReplyNVT( buf.data, buf.nUsed );
         AccumReply( " " );
         AccumReplyPathname( szPath, pathlen );
         AccumReply( "\r\n" );
      }
      else
         AccumReply( szNoMem  );
      DestroyXBuf( &buf );
   }
   else
      AccumReply( szNoMem );
   
   return SendReplies();
}
