/*
	parser.c

	Rudimentary command line parser. 

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: parser.c,v 1.19 2008/04/30 23:14:20 shesmax Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "parser.h"
#include "pi.h"
#include "dtp.h"
#include "mlsx.h"          /* Mlsd(), Mlst() */
#include "retcodes.h"

static const char szEmpty[] = "";
static const char* szPrevCmd = szEmpty;	/* holds the previous command */

/* Standard, deprecated and non-standard (but widely used) FTP commands */
static Keyword keywords[] =
{
	{"ABOR", &Abor, 0},
	{"ACCT", &Acct, 0},   
	{"ALLO", &Allo, 0},	
	{"APPE", &Appe, "<pathname>\tAppend to a file."},  
	{"CDUP", &CdUp, "Change to parent of current working directory"},
	{"CWD" , &Cwd,  "<directory>\tChange working directory to <directory>"},
	{"DELE", &Dele, "<file>\tDelete the file <file>"},
	{
		"EPRT", &EPrt,  
		"<port>\tDirects the server to open data connection to the port"
	}, 
	{
      "EPSV", &EPsv, 
      "[<all|af]\tEnter extended passive mode.\r\n"
      " ALL - Enter ext. passive mode and accept further EPSV only.\r\n"
      " af  - Specifies address family for server's data connection.\r\n"
      "       Possible values are 1 (IPv4) and 2 (IPv6).\r\n"
      "214 "
   },
	{"FEAT", &Feat, "\tList features supported by the server"},
	{
		"HELP", &Help, 
		"[<command>]\tGive list of all commands or describe the given command"
	},
   {"LANG", &Lang, "[EN]\tSets language of server\'s messages (only English is supported)"}, 
	{"LIST", &List, "[<pathname>]\tList files that match to <pathname>"}, 
	{"MDTM", &Mdtm, "<file>\tShow last modification time of the file <file>"},
	{"MKD" , &Mkd,  "<directory>\tMake the directory <directory>"},
	{
      "MLSD" , &Mlsd,  
      "[<directory>]\tList the content of <directory>, current by default"
   },
	{
      "MLST" , &Mlst,  
      "[<path>]\tList the information about <path>, current dir by default"
   },
	{"MODE", &Mode, "<mode>\tSpecify data transfer mode <mode>"},
	{"NLST", &Nlst, "[<pathname>]\tList file names that match to <pathname>"},  
	{"NOOP", &Noop, "Do nothing"},
	{"OPTS", &Opts, "Specify the desired behavior of the server"},
	{"PASS", &Pass, 0},
	{"PASV", &Pasv, "Enter passive mode"},
	{
		"PORT", &Port,  
		"<port>\tDirects the server to open data connection to the port"
	}, 
	{"PWD" , &Pwd,  "Print the current working directory"},
	{"QUIT", &Quit, "Terminate session"},
	{"REIN", &Rein, "Restore state of the connection to default values"},
	{ 
		"REST", &Rest,  
		"<bytes>\tRestart incomplete transfer at the given byte offset"
	},
	{"RETR", &Retr, "<file>\tRetrieve the file <file>"},
	{"RMD", &Rmd,   "<directory>\tRemove the directory <directory>"},
	{"RNFR", &Rnfr, "<file>\tSpecify rename-from file name"},
	{
		"RNTO", &Rnto, 
		"<file>\tSpecify rename-to file name and rename the file"
	},
	{
		"SITE", &Site, 
		"Non-standard commands\r\n"
		"SITE CHECKSUM [on|off]  \tToggle the checksum computation mode(MD5)\r\n"
		"SITE CHMOD <mode> <file>\tChange mode of the file <file> to <mode>\r\n"
      "SITE IDLE  [<secs>]     \tSet or query idle-timer\r\n"
      "SITE NEWER <date> <path>\tList files newer than a particular date\r\n"
		"                        \tand that match to <path>\r\n" 
		"                        \tThe format of <date>: YYYYMMDDHHMMSS\r\n"
		"SITE MINFO <date> <path>\tSame as SITE NEWER\r\n"
      "SITE UMASK [<mask>]     \tChange or query umask\r\n"
      "SITE HELP               \tInformation about SITE commands\r\n"
		"214 "
	},
	{"SIZE", &Size, "<file>\tReturn size of the file <file>"},
	{"SMNT", &Smnt, 0},  
	{
		"STAT", &Stat, 
		"[<path]\tReturn status of server or list <path>"
	},
	{"STOR", &Stor, "<file>\tStore the file <file>"},
	{"STOU", &Stou, "Store a file with an unique name"},
	{"STRU", &Stru, "<code>\tSpecify data transfer structure"},
	{"SYST", &Syst, "Show operating system type of server system"},
	{"TYPE", &Type, "<type>\tSpecify data transfer type <type>"},
	{"USER", &User, 0},
	{
		"XCUP", &CdUp, 
		"Change to parent of current working directory (deprecated)"
	},
	{
		"XCWD", &Cwd, 
		"<directory>\tChange working directory to <directory> (deprecated)"
 	},
	{
		"XMKD", &Mkd,  
		"<directory>\tMake the directory <directory> (deprecated)"
	},
	{"XPWD", &Pwd,  "Print current working directory (deprecated)" },
	{
		"XRMD", &Rmd,  
		"<directory>\tRemove the directory <directory> (deprecated)"
	}
};
#define DIM( a ) (sizeof a / sizeof a[ 0 ])

/*
* Parser
*/

static int CmpKwd( const void* cmd, const void* kwd )
{
	return strcasecmp( (const char*)cmd, ((const Keyword*)kwd)->szCmd );
}

Keyword* ResolveCommand( Keyword* kwds, int nKwds, const char* sCmd )
{
	return (Keyword*)bsearch( sCmd, kwds, nKwds, sizeof kwds[ 0 ], CmpKwd );
}

#include "utf8.h"
#include "log.h"
int GetPath( char** args, char** path, int bOpt )
{
   size_t len;
   
	*path = *args;
	if( !**path )
   {
      if( !bOpt )
	   {
		   AccumReply( "501 Need path specifier for this command\r\n" );
		   return RC_FALSE;
	   }
   }
   else 
   {
   	*args += (len = strlen( *args ));
   
      if( utf8_enctype( (const unsigned char *)*path, len ) == ENCTYPE_LOCAL ) 
         LogEvent( LLOG_WARN, "Received pathname in local encoding. This may "
               "lead to unpredictable results\r\n" );
   }
	return RC_TRUE;
}

static int Cvt( char* s, char* e, int low, int high )
{
	int n = 0;

	while( s < e )
	{
		if( !isdigit( *s ) )
			return -1;
		n = n * 10 + (*s++ - '0');
	}
	return n >= low && n <= high ? n : -1;
}

int GetDate( char** args, time_t* pTime )
{
	struct tm tm;

	if
	( 
		(tm.tm_year = Cvt( *args, *args + 4, 0, 9999 )) != -1 &&
		(tm.tm_mon = Cvt( *args + 4, *args + 6, 1, 12 )) != -1 &&
		(tm.tm_mday = Cvt( *args + 6, *args + 8, 1, 31 )) != -1 &&
		(tm.tm_hour = Cvt( *args + 8, *args + 10, 0, 23 )) != -1 &&
		(tm.tm_min = Cvt( *args + 10, *args + 12, 0, 59 )) != -1 &&
		(tm.tm_sec = Cvt( *args + 12, *args + 14, 0, 61 )) != -1 
	)
	{
		tm.tm_year -= 1900;
		--tm.tm_mon;
		if( (*pTime = mktime( &tm )) != -1 )
		{
			*args += 14;
			return RC_TRUE;
		}
	}
	AccumReply( "501 The date should have the format YYYYMMDDHHMMSS\r\n" );
	return RC_FALSE;
}

int ExtraParameters( char* args )
{
	if( *args )
	{
		AccumReply( "501 Syntax error - unexpected extra parameter(s).\r\n" );
		return RC_TRUE;
  	}
	return RC_FALSE;
}

int TestPrevCommand( const char* szChkCmd, const char* szCurCmd )
{
	if( strcasecmp( szPrevCmd, szChkCmd ) )
	{
		AccumReplyF( "503 Send %s before %s\r\n", szChkCmd, szCurCmd );
		return RC_FALSE;
	}
	return RC_TRUE;
}

int ExecCommand( char* szCmdLine )
{
	char* cmd = szCmdLine;
	char* args = strchr( cmd, ' ' );
	Keyword* pKwd;

	if( args )
		*args++ = '\0';	/* separate the command from its arguments */
	else
		/* Point args to an empty string. This lets to avoid further checks
			whether the passed pointer is a valid one. */
		args = cmd + strlen( cmd );

   pKwd = ResolveCommand( keywords, DIM( keywords ), cmd );

	if( pKwd )
	{
	  	int result = pKwd->handler( args );

		szPrevCmd = pKwd->szCmd;
		return result;
	}
	szPrevCmd = szEmpty;
	return SendReply( "500 Unrecognized command\r\n" );
}

/*
* FTP command handlers
*/

/* 
	Sends the help either about all commands or about the specified command
  	to the remote user.
*/
int Help( char* szCmd )
{
   unsigned int i;
	Keyword* pKwd;

	int nCol = 0;

	if( *szCmd )
	{
		if( (pKwd = ResolveCommand( keywords, DIM( keywords ), szCmd )) != 0 )
         return SendReplyF
			( 
				"214%c%-4.4s %s\r\n", 
				pKwd->szHelp && strchr( pKwd->szHelp, '\n' ) ? '-' : ' ', 
         	pKwd->szCmd, 
				pKwd->szHelp ? pKwd->szHelp : "Superfluous at this site" 
			);
		else
         return SendReply( "214 Command is not recognized\r\n" );
	}

   /* List all implemented commands */
	AccumReply( "211-Following commands are accepted by this server:\r\n" );
	for( i = 0, pKwd = keywords; i < DIM( keywords ); ++i, ++pKwd )
	{
		if( nCol >= 75 )
		{
			AccumReply( "\r\n" );
			nCol = 0;
		}
		AccumReplyF( "%-4.4s%c   ", pKwd->szCmd, pKwd->szHelp ? ' ' : '*' ); 
		nCol += 10;
	}
	if( nCol )
		AccumReply( "\r\n" );
	return SendReply
	( 
		"* - Superfluous commands, MLST and MLSD commands are experimental.\r\n"
		"211 \r\n" 
	);
}
