/*
	parser.h

	FTP commands parsing

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: parser.h,v 1.10 2008/04/30 23:14:20 shesmax Exp $ */

#ifndef __GUARD_PARSER_H
#define __GUARD_PARSER_H

/* FTP command */
typedef struct tagKeyword
{
	const char* szCmd;				/* the command  itself */
   int (*handler)( char* args );	/* command handler */
	const char* szHelp;				/* Help hint for human reader */
} Keyword;

/*
* Functions for parsing 
*/

/* 
	Resolves the passed command based on the keyword table passed in kwds.

	Parameters
	
		kwds		Keyword table.
		nSlots	Number of entries in the table
		szCmd		The command we want to resolve. Case insensitive.

	Returns 
	
		The pointer to an entry in the keyword table if the command is valid or 0.
*/
Keyword* ResolveCommand( Keyword* kwds, int nSlots, const char* szCmd );

/*
	Retrieves the path from the passed command line. Advances
	args to next parameter in command line. 

	Parameters
		
		args	Command argument(s).
		path	Receives the resulting path.
		bOpt	1 means that the path argument is optional. If it is optional
				then **args should be '\0'

 Return value
	
		Returns the pointer to the null-terminated string in path.
		Advances args so, that it will point to next parameter
		in the command line (but see the note below). Returns 1 if the
		path argument was okay, 0 otherwise. If there was a parsing error
		places the 501 reply in the reply buffer.

 Note

 Under path we understand the string of any characters except
 the terminating null and CRLF at the end of the line. According
 to this convention path can be only the last parameter in the command line.
 In that case after the execution of this function args will point to '\0'.
*/
#define GP_OPTIONAL		1			/* the path argument is optional */
#define GP_OBLIGATORY	0			/* path argument should be present */
int GetPath( char** args, char** path, int bOpt );

/*
	Retrieves the date from the command line. The date should be given
	in the following format: YYYYMMDDHHDDSS.

	Parameters
		
		args	Command line argument(s).
		date	Receives the result.

 Return value
	
		Returns the date in *date. Advances args so, that it will point to 
		next parameter. Returns 1 if the passed date was okay, 0 otherwise.
		If there was a parsing error, it places the 501 reply in the reply 
		buffer.
*/	
#ifndef _TIME_H
#	include <time.h>
#endif
int GetDate( char** args, time_t* date );

/*
	Returns 1 if there are unexpected extra parameters in the command line.
	If there is some extra text in the command line places the reply
	"501 Unexpected extra parameter(s)" in the reply buffer.
*/
int ExtraParameters( char* args_ );

/* 
	Support for command sequences. Tests whether the remote client sent
	the command szChkCmd prior to szCurCmd. If not places the corres-
	ponding "503 ..." reply in the reply buffer.
*/
int TestPrevCommand( const char* szChkCmd, const char* szCurCmd );

/* Parses and executes the given command. */
int ExecCommand( char* szCmdLine );

/* Valid numbers in PORT command are between 0 and 255 */
#define IS_PORT_NUMBER( n ) (0 <= (n) && (n) <= 255)

/* For PATH_MAX */
#include <limits.h> 
/* SITE MINFO or SITE NEWER */
#define MAXCMDVERB   10
/* <date> + space + <path> in SITE NEWER */
#define MAXARG       16 + 1 + PATH_MAX
/* 1 for space between command verb and argument; 2 for \r\n */
#define MAXCMDLEN    MAXCMDVERB + MAXARG + 1 + 2

#endif
