/* 
   pi.c

   Protocol Interprenter.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: pi.c,v 1.44 2008/04/30 23:14:20 shesmax Exp $ */

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/utsname.h>

#include <netinet/in.h> /* Internet related stuff */
#include <netinet/ip.h>
#include <arpa/inet.h>  /* for inet_ntoa */
#include <arpa/telnet.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

#include "libra.h"
#include "pi.h"
#include "dtp.h"
#include "parser.h"
#include "retcodes.h"
#include "listdir.h"
#include "mlsx.h"       /* MlsxListFacts */
#include "sockets.h"
#include "log.h"
#include "xbuf.h"

static const char sVersion[] = "Libra FTP daemon"
#ifdef LIBRA_VERSION
" " LIBRA_VERSION
#endif
;

/* Requests & Replies */
static XBuf xbReply;    /* module static data - used in several functions */
static XBuf xbCmd;      /* declared static to reduce number of alloc/dealloc */
static XBuf xbCmdAccum; /* command accumulation buffer - has to be static */

static int nUTF8_Ctrl = RC_FALSE;
static int nUTF8_Data = RC_FALSE;
static int nClientIsMSIE = RC_FALSE;
/*
* Out-of-band data
*/

static char* ReadCmdOob( void );
static void OobData()   /* SIGURG handler */
{
   char* sCmd;

   if( !DataTransferInProgress() )
      return;  /* Do not bother */
   
   if( (sCmd = ReadCmdOob()) == 0 )
      return;  /* The control connection is closed */

   if( strcasecmp( sCmd, "ABOR" ) == 0 )
      PostTransferAbort();
   else if( strcasecmp( sCmd, "STAT" ) == 0 )
      SendReplyF( "213 Sent %ld of %ld bytes\r\n", DataSent(), DataToSend() );
}

/*********************************************************************
*  Initailization/Deinitialization
*/

static int PIInitBuffers( void )
{
   int n;

   n = 0;
   do 
   {
      if( CreateFixedXBuf( &xbReply, REPLY_BUFSIZE ) == XBUF_OVERFLOW )
         break;
      n++;
      if( CreateFixedXBuf( &xbCmd, MAXCMDLEN ) == XBUF_OVERFLOW )
         break;
      n++;
      if( CreateFixedXBuf( &xbCmdAccum, MAXCMDLEN ) == XBUF_OVERFLOW )
         break;
      n++;
   } 
   while( 0 );

   switch( n )
   {
   case 3:
      return 1;
   case 2:
      DestroyXBuf( &xbCmd );
      /* fall through */
   case 1:
      DestroyXBuf( &xbReply );
      /* fall through */
   default:
      return 0;
   }
}

static void PIDestroyBuffers( void )
{
   DestroyXBuf( &xbReply );
   DestroyXBuf( &xbCmd );
   DestroyXBuf( &xbCmdAccum );
}

static int InitSigHandlers( void )
{
   struct sigaction sa;

   /* Out-of-band data */
   sa.sa_handler = &OobData;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = SA_RESTART;
   if( sigaction( SIGURG, &sa, 0 ) == -1 )
      LogEvent
      ( 
         LLOG_WARN, "Failed to setup handler for out-of-band data: %s\n",
         strerror( errno )
      );
   return 1;
}

/*
*  Control connection addresses 
*/
PSaddr addrPeerCtrl = 0;
PSaddr addrSrvrCtrl = 0;

/*
* Setup the contol connection.
*/

static int CtrlSocketInit( void )
{
   socklen_t socklen;
   int x;

   /* Get the address of the control connection. We need it
      because we should open the data connection on the same
      IP address.
   */
   socklen = sizeof( Saddr );
   if( getsockname( STDIN_FILENO, addrSrvrCtrl, &socklen ) == -1 )
   {
      LogEvent( LLOG_ERROR, "getsockname: %s\n", strerror( errno ) );
      return 0;
   }
   if( !AF_Supported( addrSrvrCtrl->sa_family ) )
   {
      LogEvent
      ( 
         LLOG_ERROR, "Bad address family %d of server end\n", 
         addrSrvrCtrl->sa_family 
      );
      return 0;
   }

   socklen = sizeof( Saddr );
   if( getpeername( STDIN_FILENO, addrPeerCtrl, &socklen ) == -1 )
   {
      LogEvent( LLOG_ERROR, "getpeername: %s\n", strerror( errno ) );
      return 0;
   }
   if( !AF_Supported( addrPeerCtrl->sa_family ) )
   {
      LogEvent
      ( 
         LLOG_ERROR, "Bad address family %d of peer end\n", 
         addrPeerCtrl->sa_family 
      );
      return 0;
   }

   if
   (
      /* Timeout handling */
      (x = fcntl( STDIN_FILENO, F_GETFL )) == -1 ||
      fcntl( STDIN_FILENO, F_SETFL, x | O_NONBLOCK ) == -1 ||
      (x = fcntl( STDOUT_FILENO, F_GETFL )) == -1 ||
      fcntl( STDOUT_FILENO, F_SETFL, x | O_NONBLOCK ) == -1
   )
   {
      LogEvent( LLOG_WARN, "fcntl(F_[GS]ETFL): %s\n", strerror( errno ) );
      return 0;
   }

   /* To receive the SIGURG signal */
   if( fcntl( STDOUT_FILENO, F_SETOWN, getpid() ) == -1 )
      LogEvent( LLOG_WARN, "fcntl(2, F_SETOWN): %s\n", strerror( errno ) );

   /* IP option "Better latency" for control connection */
   x = IPTOS_LOWDELAY;
   if
   ( 
      setsockopt( STDIN_FILENO, IPPROTO_IP, IP_TOS, &x, sizeof x ) == -1 ||
      (
         x = IPTOS_LOWDELAY, 
         setsockopt( STDOUT_FILENO, IPPROTO_IP, IP_TOS, &x, sizeof x ) == -1 
      )
   )
      LogEvent( LLOG_WARN, "IPTOS_LOWDELAY: %s\n", strerror( errno ) );

   x = 1;
   if( setsockopt( STDIN_FILENO, SOL_SOCKET, SO_OOBINLINE, &x, sizeof x ) < 0 )
      LogEvent
      ( 
         LLOG_WARN, "Failed to setup inline out-of-band data: %s\n",
         strerror( errno )
      );

   return 1;
}

#define MAX_ADDR 80
static char *PartyName( PSaddr addr, char* envvar )
{
   char* szName;
   char* s;
   
   /* Try to detect the name from tcpserver first */
   s = getenv( envvar );
   if( !s )
   {
      if( (szName = (char*)malloc( MAX_ADDR )) != 0 )
         if( SockAddrString( addr, szName, MAX_ADDR ) == -1 )
            strcpy( szName, "???" );
   }
   else
   {
      size_t n = strlen( s ) + 1;
      
      if( (szName = malloc( n )) != 0 )
         memcpy( szName, s, n );
   }
   return szName;
}

/*
* Initialize PI module.
* Return TRUE on success, FALSE on fail.
*/
int InitPI()
{
   static char szErr[] = "421 Cannot initialize the control connection\r\n";
   int buffersInitialized;

   buffersInitialized = 0;
   do {
      if( (addrPeerCtrl = SockAddrCreate()) == 0 )
         break;
      if( (addrSrvrCtrl = SockAddrCreate()) == 0 )
         break;
      if (!PIInitBuffers())
         break;
      buffersInitialized = 1;
      if (!InitSigHandlers())
         break;
      if (!CtrlSocketInit())
         break;
      return 1;
   } while( 0 );

   if( buffersInitialized )
      PIDestroyBuffers();

   /* Send the refusal to the peer but without any buffering */
   write( STDOUT_FILENO, szErr, sizeof szErr - 1 );
   return 0;
}

/*
* Exit routine for protocol interprenter.
* Called only if InitPI has succeeded.
*/
void ShutdownPI()
{
   PIDestroyBuffers();
   SockAddrDestroy( addrPeerCtrl );
   SockAddrDestroy( addrSrvrCtrl );
   shutdown( STDIN_FILENO, 2 );
}

/*********************************************************************
   Read a command from the remote client. 
   
   RFC 1123 requirements:
   -  The command can be of any size.
   -  There can be more than one command in the received TCP segment.
   -  FTP should abandon any Telnet options negotiation.

   But!
   The OS defines the limit of path name, therefore actually, we can abandon 
   any command longer then maximum path name length plus longest command plus 
   longest argument plus delimiters and \r\n.

   To stay in sync with client we should read the input until we see \r\n.
*/

/* Process Telnet commands */
static void TelnetProcess( XBuf* xbuf )
{
   unsigned char* p;
   unsigned char* e;
   unsigned char* s;

   /* 
      Set e -> \r. We put \0 there intentionaly - this is used in algorithm 
      of parsing Telnet commands.
   */
   e = (unsigned char*)xbuf->data + xbuf->nUsed - 1;   
   *e = '\0';   
   for( p = s = (unsigned char*)xbuf->data; p < e; ++p )
   {
      if( *p == IAC )   /* Interpret As Command */
      {
         switch( p[ 1 ] )
         {
         case DO:
            if( p + 2 < e )
            {
               /* Answer to remote: IAC WONT <option code> */
               p[ 1 ] = WONT;
               while( write( STDOUT_FILENO, p, 3 ) == -1 && errno == EINTR )
                  ;
            }
            /* else violation of Telnet IAC DO \0 */
            p += 2;
            continue;

         case WILL:
            if( p + 2 < e )
            {
               /* Answer to remote: IAC DONT <option code> */
               p[ 1 ] = DONT; 
               while( write( STDOUT_FILENO, p, 3 ) == -1 && errno == EINTR )
                  ;
            }
            /* else violation of Telnet IAC WILL \0 */
            p += 2;
            continue;

         /* 
            I am not sure we can get following two commands because they
            are answers to option negotiation, what we won't initiate.
         */
         case DONT:  
         case WONT: 
            p += 2;  /* skip three characters in command stream */
            continue;

         case IAC:   /* Escaped IAC */
            ++p;     /* copy IAC into the command */
            break;
         default: 
            /* 
               Here we can have in p[ 1 ] either one of the remaining
               Telnet commands (NOP, DM, IP, AYT etc) or some kind of
               violation of Telnet stream - IAC followed by some 
               "normal" character. I would prefer to skip this sequence.
            */
            ++p;
            continue;
         }
      }

      /* *p is some character of command request */
      if( s != p )
         *s = *p;
      ++s;
   }
   *s = '\0';
}

static int GetCommand( XBuf* dst, XBuf* src, int len )
{
   /* copy the command and \r to the command buffer */
   ClearXBuf( dst );
   if( WriteXBuf( dst, src->data, len - 1 ) == XBUF_OVERFLOW )
      return 0;   
   /* remove the head of the accumulating buffer */
   LTrimXBuf( src, len );
   return 1;
}

static int SkipRestOfCommand()
{
   int nBytes;
   int nLineLen;

   ClearXBuf( &xbCmdAccum );
   for(;;)
   {
      nBytes = ReadSockXBuf( &xbCmdAccum, STDIN_FILENO );
      if( nBytes <= 0 )
         return RC_FALSE;
      nLineLen = LineLenXBuf( &xbCmdAccum );
      if( nLineLen != 0 )
      {
         LTrimXBuf( &xbCmdAccum, nLineLen );
         return RC_TRUE;
      }
   }
}
         
static char* ReadCmd( void )
{
   int nBytes, nLineLen;
  
   /* Read the command either from the accumulated buffer or from the socket */
   while( ( nLineLen = LineLenXBuf( &xbCmdAccum ) ) == 0 )
   {
      if( FreeSpaceXBuf( &xbCmdAccum ) == 0 )
      {
         if( !SkipRestOfCommand() )
            return 0;
         if( SendReply( "501 Command line too long\r\n" ) != RC_OK )
            return 0;
      }
      else
      {
         nBytes = ReadSockXBuf( &xbCmdAccum, STDIN_FILENO );
         if (nBytes <= 0)
            return 0;
      }
   }

   if( !GetCommand( &xbCmd, &xbCmdAccum, nLineLen ) )
      return 0;
   TelnetProcess( &xbCmd );
   LogEvent( LLOG_CMD, "%s\n", xbCmd.data );
   
   return xbCmd.data;
}

static char* ReadCmdOob( void )
{
   /* FIXME */
   return ReadCmd();
}

/*********************************************************************
   Upper-level PI functions.
*/

static int Welcome()
{
   char* nodename;
   struct utsname uts;
   char* buf;

   /* 
      Get the name of the node either from environment (for virtual
      hosts) or system wide.
    */
   nodename = getenv( "TCPLOCALHOST" ); /* tcp-server specific */
   if( !nodename )
      if( uname( &uts ) != -1 )
         nodename = uts.nodename;
   if( !nodename )
      nodename = "";

   buf = (char*)malloc( strlen( nodename ) + strlen( sVersion ) + 4 );
   if( buf )
   {
      int r;
      
      sprintf( buf, "%s (%s)", nodename, sVersion );
      r = DirInfo( 220, buf );
      free( buf );
      return r;
   }
   return SendReply( "220 Welcome!\r\n" );
}

/* Communicates with peer by reading its requests and executing them. */
int RunPI()
{
   char* sPeer = PartyName( addrPeerCtrl, "TCPREMOTEHOST" );

   LogEvent( LLOG_CONNECT, "%s\n", sPeer );
   free( sPeer );
   if( Backend() || Welcome() == RC_OK )
      for(;;)
      {
         char* sCmd;

         if( (sCmd = ReadCmd()) == 0 || ExecCommand( sCmd ) != RC_OK )
            break;
      }
   return RC_OK;
}

/*
* Command handlers
*/

int Abor( char* unused UNUSED_PARAM )
{
   return SendReply( "226 No open data connection\r\n" );
}

/* Returns the type of the operating system used at the server's site. */
int Syst( char* args )
{
   if( ExtraParameters( args ) )
      return SendReplies();
   return SendReply( "215 UNIX Type: L8\r\n" );
}

/* Auxiliary command for STAT * */
static int LongListEntry( const char* name, const struct stat* st, void* nFnd )
{
   char sPrefix[ 57 ];

   if( ++(*(int*)nFnd) == 1 )
      AccumReply( "212-\r\n" );
   LsLongPrefix( st, sPrefix, sizeof sPrefix );
   AccumReply( sPrefix );
   AccumReplyPathname( name, strlen( name ) ); 
   return AccumReply( "\r\n" );     
}

/* Status response */
int Stat( char* args )
{
   static char sOn[] = "On";
   static char sOff[] = "Off";
  
   if( !*args )      /* Return the information about the server */
   {
      char* sHost = PartyName( addrSrvrCtrl, "TCPLOCALHOST" );
      char* sPeer = PartyName( addrPeerCtrl, "TCPREMOTEHOST" );
      char sPeerData[ MAX_ADDR ], sSrvrData[ MAX_ADDR ];
      static char sQst[] = "???";

      if( SockAddrString( addrPeerData, sPeerData, sizeof sPeerData ) == -1 )
         strcpy( sPeerData, sQst );
      if( SockAddrString( addrSrvrData, sSrvrData, sizeof sSrvrData ) == -1 )
         strcpy( sSrvrData, sQst );
      AccumReplyF
      (
         "211- Server Status\r\n"
         " Control connection:\r\n"
	 		"  Server=%s, Client: %s, Command buffer: %d bytes, Reply buffer: %d bytes\r\n"
         " Data connection (%s):\r\n"
	 		"  Server: %s Client: %s,\r\n"
         "  Type: %s, Structure: %s, Mode: %s,\r\n"
         "  Checksums: %s, Passive only: %s, EPSV only: %s%s\r\n"
         "211 \r\n",
         sHost ? sHost : sQst, sPeer ? sPeer : sQst, xbCmd.nSize, xbReply.nSize,
         GetDataConnType(), 
         sSrvrData, sPeerData,           
         GetRepresentationType(), GetFileStructure(), GetTransferMode(),
         ComputeChecksums() ? sOn : sOff, PassiveOnlyMode() ? sOn : sOff, 
         EpsvOnly() ? sOn : sOff,
         ClientIsMSIE() ? ", Client: MSIE" : ""
      );
      free( sHost );
      free( sPeer );
   }
   else     /* STAT * */
   {
      int nEntries = 0;
		
      if( Ls( args, LongListEntry, (void*)&nEntries ) == RC_OK )
         AccumReplyF( "212 %d entries\r\n", nEntries );
      else
         PError( 212 );
   }
   return SendReplies();
}

/* Sends OK reply to the peer. */
int Noop( char* unused UNUSED_PARAM )
{
   return SendReply( "200 OK\r\n" );
}

/*
* RFC 2389 - Feature negotiation mechanism for the File Transfer Protocol
*/

/* Lists experimental and non-standard commands supported at this moment. */
int Feat( char* args )
{
   if( !ExtraParameters( args ) )
   {
      AccumReply( "211-Extensions supported\r\n" );
      
      AccumReply
      (
         " MDTM\r\n"
         " REST STREAM\r\n"
         " SIZE\r\n"
         " UTF8\r\n"
         " TVFS\r\n"
         " LANG EN*\r\n"
         " SITE CHECKSUM\r\n"
         " SITE CHMOD\r\n"
         " SITE IDLE\r\n"
         " SITE MINFO\r\n"
         " SITE NEWER\r\n"
         " SITE UMASK\r\n"
      );

      AccumReply( " MLST" );
      MlsxListFacts( MLF_ALL_FACTS );
      AccumReply( "\r\n" ); 

      AccumReply( "211 END\r\n" );
   }
   return SendReplies();
}

/* Allows to set server specific options. */  
int Opts( char* args )
{
   static char szMlst[] = "MLST";
   static char szUTF8_PLAIN[] = "UTF8";
   static char szUTF8_NLST[] = "UTF8 NLST";
   static char szUTF8_MSIE[] = "UTF8 ON";
   static char szReply0[] = "200 Pathnames over control and data connections will be sent in UTF-8\r\n";
  
   if( strncasecmp( args, szMlst, sizeof szMlst - 1 ) == 0 )
      return MlsxOpts( args + sizeof szMlst - 1 );
   
   if( !strcasecmp( args, szUTF8_PLAIN ) ) {
      nUTF8_Ctrl = RC_TRUE;
      return SendReply( "200 Pathnames over control connection will be sent in UTF-8\r\n" );
   }
   if( !strcasecmp( args, szUTF8_NLST ) ) {
      nUTF8_Ctrl = RC_TRUE;
      nUTF8_Data = RC_TRUE;
      return SendReply( szReply0 );
   }

   // MS IE sends opts utf8 on. This is not specified in any
   // standart, but MS always does what it wants.
   if( !strcasecmp( args, szUTF8_MSIE ) ) {
      nUTF8_Ctrl = RC_TRUE;
      nUTF8_Data = RC_TRUE;
      nClientIsMSIE = RC_TRUE;
      return SendReply( szReply0 );
   }
   return SendReply( "501 Invalid syntax\r\n" );
}

int UTF8PathnamesOverCtrlConnection()
{
   return nUTF8_Ctrl;
}

int UTF8PathnamesOverDataConnection()
{
   return nUTF8_Ctrl;
}

int ClientIsMSIE()
{
   return nClientIsMSIE;
}

void ResetUTF8Usage()
{
   nUTF8_Ctrl = nUTF8_Data = nClientIsMSIE = RC_FALSE;
}

int Lang( char* args )
{
   static char szEN[] = "EN";
   
   if (!*args || !strncasecmp( args, szEN, sizeof szEN - 1) )
      return SendReply( "200 English will be used for all messages from the server\r\n" );
   return SendReply( "501 Invalid language tag(s)\r\n");
}

/*
* Replies to the peer
*/

/* 
   Last Reply Status - holds the status of the serie of ...Reply... calls. 
   This simplifies the code, because it avoids the necessity to check the
   return code of ...Reply... functions all the time. You write the program as
   if they all can't fail and check the return code only of the last one.
*/
static int nLRS = RC_OK;

/* 
   This flag is necessary for the implementation of reply audit. 
   We can't write a reply from the server to the audit log at the time
   of buffer flush because probably we'll do it several times for a long reply.
   It would be very unpleasant for the reader to find one part of the 
   answer at line number X and the rest of this answer several hundred
   lines below (on heavy used server ;-).
   Hence the program writes only the start of an answer to the audit log.
*/
static int bStartOfReply = 1;

static int FlushReplyBuf( XBuf* buf, int socket )
{
   if( bStartOfReply )
   {
      /* Log output will incorrectly work if there are some \0 in the buffer. It
       * happens, for example, if we send a file name with \r in it.
       * This is considered as minor issue. It is not worth of effort to
       * display \0 in a printable manner */
      LogEvent( LLOG_REPLY, "%.*s", buf->nUsed, buf->data );
      bStartOfReply = 0;
   }
   return FlushSockXBuf( buf, socket );
}

int AccumReplyBuf( const char* text, size_t sz )
{
   if( nLRS == RC_OK )
      nLRS = PutSockXBuf( &xbReply, text, sz, FlushReplyBuf, STDOUT_FILENO );
   return nLRS;
}

#include "utf8.h"
static char escaped_iac[ 2 ] = {IAC, IAC};
static char escaped_cr[ 2 ] = {'\r', '\0'};
   
/* This function sends data over control connection using following rules:
 *    1) Telnet's Interpret As Command (IAC == \xFF) is sent as IAC IAC;
 *    2) \r alone is sent as \r\0;
 *    3) other characters are sent as UTF-8.
 */
int AccumReplyPathname( const char* path, size_t sz )
{
	const unsigned char* start = (const unsigned char*)path;	
	const unsigned char* end = (const unsigned char*)path + sz;

   if( nLRS != RC_OK )
   	return nLRS;

	while( start < end )
	{
		const unsigned char* p = start;
      size_t len;

      /* Search for the presence of IAC or \r alone */
		while( p < end && *p != IAC && *p != '\r' ) 
			++p;

      /* If UTF-8 encoded pathnames are required for the data connection then we
       * need to convert pathnames that are coming from filesystem to UTF-8 
       * prior to send them to the User DTP. Such conversion is only necessary 
       * if a pathname contains characters from a local character set. ASCII 
       * characters can be transmitted without any conversion.
       *
       * If UTF-8 encoding of pathnames is not required we send pathnames as is.
       * This behavior violates the requirement of RFC959 to send only ASCII 
       * characters in pathnames. It was intentionally programmed so in order to
       * provide the possibility to transmit files with language-specific names.
       * This is working well of cause  only if a client and the server use  
       * the very same character set. 
       *
       * If filesystem is using UTF-8 encoded pathnames then they are also sent 
       * without any modification to the client program. This was done so because 
       * the server has no way to determine the character set of the client and 
       * thus cannot do any meaningfull conversion of pathnames. One option would
       * be to replace all non-ASCII characters with some ASCII character, but
       * this would give no chance to the client to convert a pathname properly. */
      len = p - start;
      if (utf8_enctype( start, len ) == ENCTYPE_LOCAL && nUTF8_Ctrl == RC_TRUE )
      {
         unsigned char* utf8;

         if( !(len = local_to_utf8( start, len, &utf8)) )
         {
            nLRS = RC_ERROR;
            break;
         }
         nLRS = PutSockXBuf( &xbReply, (char*)utf8, len, FlushReplyBuf, STDOUT_FILENO );
         free( utf8 );
         if( nLRS != RC_OK )
            break;
      }
      else if( (nLRS = PutSockXBuf( &xbReply, (char*)start, len, FlushReplyBuf, STDOUT_FILENO )) != RC_OK )
			break;
      
      if( p == end )
			break;         /* end of pathname reached */
      
      /* if special characters are found, send them escaped */
      nLRS = PutSockXBuf( &xbReply, *p == IAC ? escaped_iac : escaped_cr, 2, FlushReplyBuf, STDOUT_FILENO );
		if( nLRS != RC_OK )
			break;
      
      start = p + 1;
	}
   return nLRS;
}

/* Special characters are escaped before sending them via control connection:
 *    1) Telnet's Interpret As Command (IAC == \xFF) is sent as IAC IAC
 *    2) \r alone is sent as \r\0
 */
int AccumReplyNVT( const char* text, size_t sz )
{
	const unsigned char *start = (const unsigned char*)text;	
	const unsigned char* end = (const unsigned char*)text + sz;

   if( nLRS != RC_OK )
   	return nLRS;
	
	while( start < end )
	{
		const unsigned char* p = start;

      /* Search for the presence of IAC or \r alone */
		while( p < end && *p != IAC && (*p != '\r' || (p + 1 < end && p[ 1 ] == '\n')) ) 
			++p;

      /* Output text before special characters */
		if( (nLRS = PutSockXBuf( &xbReply, (const char*)start, p - start, FlushReplyBuf, STDOUT_FILENO )) != RC_OK )
			break;

      if( p == end )
			break;
      
      /* if special characters are found, send them escaped */
      nLRS = PutSockXBuf( &xbReply, *p == IAC ? escaped_iac : escaped_cr, 2, FlushReplyBuf, STDOUT_FILENO );
		if( nLRS != RC_OK )
			break;
      start = p + 1;
	}
   return nLRS;
}

static int PutReply( const char* fmt, va_list args )
{
   XBuf tmp;

   if( CreateExpandableXBuf( &tmp, 256 ) == XBUF_OVERFLOW )
      return nLRS = RC_ERROR;
   if( VPrintfXBuf( &tmp, fmt, args ) == XBUF_OVERFLOW )
      nLRS = RC_ERROR;
   else
      AccumReplyBuf( tmp.data, tmp.nUsed );
   DestroyXBuf( &tmp );
   return nLRS;
}

int AccumReplyF( const char* fmt, ... )
{
   va_list args;

   va_start( args, fmt );
   PutReply( fmt, args );
   va_end( args );
   return nLRS;
}

int SendReplyBuf( const char* text, size_t nBytes )
{
   AccumReplyBuf( text, nBytes );
   return nLRS == RC_OK ? SendReplies() : nLRS;
}

int SendReplyF( const char* fmt, ... )
{
   va_list args;

   va_start( args, fmt );
   PutReply( fmt, args );
   va_end( args );
   return nLRS == RC_OK ? SendReplies() : nLRS;
}

int SendReplies()
{
   if( bStartOfReply )
      LogEvent( LLOG_REPLY, "%.*s", (int)xbReply.nUsed, xbReply.data );
   if( nLRS == RC_OK )
      nLRS = FlushSockXBuf( &xbReply, STDOUT_FILENO );
   ClearXBuf( &xbReply );
   bStartOfReply = 1;
   return nLRS;
}

void PError( int nCode )
{
   AccumReplyF( "%d %s\r\n", nCode, strerror( errno ) );
}
