/*
	pi.h

	Declarations for Protocol Interprenter.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: pi.h,v 1.23 2008/04/30 23:14:20 shesmax Exp $ */

#ifndef __GUARD_PI_H
#define __GUARD_PI_H

#include "libra.h"

/*
*	Initialization
*/

/* Initialise the protocol interprenter. Returns 1 on success. */
int  InitPI();
/* Shuts down the interprenter. */
void ShutdownPI();
/* Main loop of the program */
int  RunPI();

/*
*	FTP command handlers
*/

int Abor( char* args );
int Acct( char* args );
int Allo( char* args );
int CdUp( char* args );
int Cwd( char* args );
int Dele( char* args );
int Feat( char* args );
int Help( char* args );
int Lang( char* agrs );
int Mdtm( char* args );
int Mkd( char* args );
int Noop( char* args );
int Opts( char* args );
int Pass( char* args );
int Pwd( char* args );
int Quit( char* args );
int Rein( char* args );
int Rmd( char* args );
int Rnfr( char* args );
int Rnto( char* args );
int Site( char* args );
int Size( char* args );
int Smnt( char* args );
int Stat( char* args );
int Syst( char* args );
int User( char* args );

/*
*	Control connection
*/

#include "sockets.h"

/* Addresses of both ends of control connection */
extern PSaddr addrPeerCtrl;
extern PSaddr addrSrvrCtrl;

/*
*	Replies
*/

#define REPLY_BUFSIZE   4096

/*
	Replies are accumulated in a reply buffer. Each time the buffer is full
	it's contents are sent to a peer.

	An accumulation of reply texts is done with the help of AccumReply... 
	functions. Send... functions write the given text to the reply buffer and
	than send it to the peer.

	For each of these two groups of functions there are three versions:
	
		XXXBuf  	Accepts pointer to a buffer with a reply text and the 
					length of the text.
		XXXF  	Allows formatted output.
		XXX  		Is actually a macro that is convinient to use if you know the 
					length of the text at compile time. This is true for most places
					of the code.
		
   AccumReplyPathname is used for sending pathnames over a control connection.
   If necessary it converts pathname to UTF-8 string prior to sent it.
   
	AccumReplyNVT should be used when you put into a reply buffer an information 
	which contents you cannot control. In this case server should properly 
   escape \r and IAC characters. You do not need to use this function for
   program replies, because 
    1) you can assure that these replies doe not contain IAC or pure \r characters. 
    2) AccumReplyNVT is slower then AccumReply.

	Please note that formated output functions create a temporary buffer to make
	a reply text and then they send this text to the reply buffer. Please use
	these functions only in cases where the length of a resulting
	text is under your control. 
	

	Return codes
 	
		RC_OK 				Text is saved in the reply buffer.
		RC_TIMEDOUT			The program attempted to send whole or part of a reply 
								text to a peer. At the end it failed to do so, because 
                        the peer did'nt respond. The contents of the reply
                        text buffer is undefined from now on. 
		RC_ERROR				The program attempted to send whole or part of a reply 
								text to a peer but it failed to do so because of some
								network or OS error. The contents of the reply text buffer 
                        is undefined from now on. 
*/
#ifndef _SIZE_T
#	include <stdarg.h>	/* for size_t */
#endif

int AccumReplyBuf( const char* buf, size_t nBytes );
int AccumReplyF( const char* fmt, ... );
#define AccumReply( text ) \
	({ typeof(text) x __attribute__((__unused__)) = {}; \
         AccumReplyBuf( text, sizeof( text ) - 1 ); })
int AccumReplyPathname( const char* path, size_t sz );
int AccumReplyNVT( const char* buf, size_t nBytes );

int SendReplyBuf( const char* text, size_t nBytes );
int SendReplyF( const char* fmt, ... );
#define SendReply( text ) \
   ({ typeof(text) x __attribute__((__unused__)) = {}; \
         SendReplyBuf( text, sizeof( text ) - 1 ); })

/*
	Sends all accumulated replies (if any) to the remote host.
	
	Returns

		RC_OK				Reply was sent to the peer.
		RC_ERROR			An error occured during the transmition of reply.
		RC_TIMEDOUT		Timeout expired during the transmition of reply.
*/
int SendReplies();

/* Accumulates the reply who consists of nCode and system error string */
void PError( int nCode );

/*
*	Miscelaneous
*/

/* 
	Sends the content of the file .message in the current directory through
	the control connection. 
	
	Parameters
	
	code	reply code 
	sPfx	prefix line (have to be sent in any case)
*/
int DirInfo( int code, const char* sPfx );

/* 
	Constructs the line which prefixes the directory entry's name in
	output from ls -l. 

	Parameters

		st				file attributes
		sPrefix		buffer for prefix
		nBytes		size of the buffer
*/
struct stat;
char* LsLongPrefix( const struct stat* st, char* sPrefix, int nBytes );


/* Returns RC_TRUE if pathnames being sent over control connection should be
   in UTF-8 */
int UTF8PathnamesOverCtrlConnection();

/* Returns RC_TRUE if pathnames being sent over data connection should be
   in UTF-8 */
int UTF8PathnamesOverDataConnection();

/* Resets the usage of UTF8 as described in UTF-8 Option for FTP */
void ResetUTF8Usage();

#endif
