/*
   port_pasv.c 
   
   FTP command handlers for PORT, PASV, EPRT, EPSV

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: port_pasv.c,v 1.11 2002/12/19 05:36:47 shesmax Exp $ */

#include <sys/types.h>

#include <unistd.h>     /* for close */
#include <stdio.h>      /* for sscanf */
#include <stdlib.h>     /* for strtoul */
#include <string.h>     /* for strlen */

#include <sys/socket.h>    /* socket related stuff */
#include <arpa/inet.h>     /* for inet_ntoa */

#include "libra.h"      /* for UNUSED_PARAM */
#include "pi.h"         /* for addrPeerCtrl */
#include "parser.h"     /* for ExtraParameters */
#include "retcodes.h"
#include "dtp.h"
#include "dataopen.h"
#include "chkport.h"
#include "sockets.h"    /* for INVALID_SOCKET */
#include "inet6_xtoy.h"

static char emsgNotImpl[] = "500 Command is not implemented\r\n";
static char emsgPsvOpen[] = "425 Can\'t open passive data connection\r\n";
static char emsgNoAddr[] = "501 No address specification\r\n";
static char emsgClntInvFam[] = 
   "425-\r\n"
   " Client has sent an %s address. Server should convert it to %s.\r\n"
   " This conversion is not possible. Data connection could not be established.\r\n"
   "425 \r\n";
static char emsgClntInvAddr[] = "425 Client sent an invalid address of data connection\r\n";
static char emsgClntInvPort[] = "501 Invalid port specification\r\n";
static char emsgSrvrInvFam[] = 
   "425-\r\n"
   " Client requested to open an %s connection.\r\n"
   " Server can open an %s connection only.\r\n"
   "425 \r\n";

/* 
* Address handling 
*/

PSaddr addrPeerData = 0;       
PSaddr addrSrvrData = 0;      

int SrvrDataConn( sa_family_t af )
{
	SockAddrAssign( addrSrvrData, addrSrvrCtrl );
   if( !SockAddrSetFamily( addrSrvrData, af ) )
	{	
		AccumReplyF( emsgSrvrInvFam, AF_Name( af ), AF_Name( addrSrvrCtrl->sa_family ) );
      SockAddrAny( addrSrvrData );
      return RC_ERROR;
	}
	SockAddrSetPort( addrSrvrData, 0 );
	return RC_OK;
}

/* 
* Connection parameters 
*/

/* RC_TRUE = passive connection */
int bPassive = RC_FALSE;                      
/* RC_TRUE = Server will reject PORT and EPRT */
static int bPassiveOnly = RC_FALSE;
/* RC_TRUE = server will reject PORT, EPRT and PASV */
static int bEpsvOnly = RC_FALSE;   

char* GetDataConnType()
{
   return bPassive ? "Passive" : "Active";
}

int PassiveOnlyMode()
{
   return bPassiveOnly;
}

void SetPassiveOnlyMode()
{
   bPassiveOnly = RC_TRUE;
}

int EpsvOnly()
{
   return bEpsvOnly;
}

/*
* So-called active mode: server opens data connection to client.
*/

/* Helper routine for Port() and Eprt() */
static int ExecPort( PSaddr addr )
{
   if( PassiveOnlyMode() || EpsvOnly() )
      return SendReply( emsgNotImpl );
   if( SrvrDataConn( addrSrvrCtrl->sa_family ) == RC_ERROR )
		return SendReplies();
   if( !CheckDataAddr( addrPeerCtrl, addr ) )
      return SendReply( emsgClntInvAddr );
   SockAddrAssign( addrPeerData, addr );
   if( !SockAddrSetFamily( addrPeerData, addrSrvrCtrl->sa_family ) )
      return SendReplyF( emsgClntInvFam, AF_Name( addr->sa_family ), AF_Name( addrSrvrCtrl->sa_family ) );
   bPassive = RC_FALSE;
   return SendReply( "200 PORT command OK\r\n" );
}

/* PORT command handler: Defines peer's data connection address */
int Port( char* args )
{
   int h1, h2, h3, h4;
   int p1, p2;
   int nChars = 0;
   struct sockaddr_in addr;
     
   /* Parse the arguments and return error code if something is ill-formed. */
   if
   (
      sscanf
      (
         args, "%d,%d,%d,%d,%d,%d%n",
         &h1, &h2, &h3, &h4,
         &p1, &p2,
         &nChars
      ) != 6 || nChars == 0 ||
      args[ nChars ] ||
      !IS_PORT_NUMBER( h1 ) || !IS_PORT_NUMBER( h2 ) ||
      !IS_PORT_NUMBER( h3 ) || !IS_PORT_NUMBER( h4 ) ||
      !IS_PORT_NUMBER( p1 ) || !IS_PORT_NUMBER( p2 )
   )
      return SendReply( emsgClntInvPort );
   /* The port specification was OK, construct peer's address. */
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = htonl
   (
      (h1 << 24) | (h2 << 16) | (h3 << 8) | h4
   );
   addr.sin_port = htons( (p1 << 8) | p2 );

   return ExecPort( (PSaddr)&addr );
}

/* Helper routine for Eprt: Reads a numeric from a address specifier */
static int GetNum( char** s, char delim, uint16_t* num )
{
   char *e;

   for( e = *s;; ++e )
      if( !*e )
         return -1;
      else if( *e == delim )
      {
         char *ep;
         
         *e = '\0';
         *num = (uint16_t)strtoul( *s, &ep, 10 );
         if( *ep )
            return -1;
         *s = e + 1;
         return 0;
      }
}

/* Helper routine for Eprt(): Separates address value */
static char *GetAddr( char* s, char delim )
{
   char *e;

   for( e = s; *e; ++e )
      if( *e == delim )
      {
         *e++ = '\0';
         return e;
      }
   return 0;
}

/* EPRT command handler: Specifies peer's data connection address */
int EPrt( char* args )
{
   struct sockaddr_in ip4;
   struct sockaddr_in6 ip6;
   uint16_t af;
   char delim = *args++;
   char *e;
   
   if( !delim )
      return SendReply( emsgNoAddr );
   if( GetNum( &args, delim, &af ) == -1 )
      return SendReply( "501 Address family is not specified\r\n" );
   e = GetAddr( args, delim );
   if( !e )
      return SendReply( emsgNoAddr );
   switch( af )
   {
   case 1:
      ip4.sin_family = AF_INET;
      if( !inet_aton( args, &ip4.sin_addr ) )
         return SendReply( "501 Syntax error in IP4 address\r\n" );
      args = e;
      if( GetNum( &args, delim, &ip4.sin_port ) == -1 )
         return SendReply( emsgClntInvPort );
      ip4.sin_port = htons( ip4.sin_port );
      break;
   case 2:
      ip6.sin6_family = AF_INET6;
      if( !inet6_aton( args, &ip6.sin6_addr ) )
         return SendReply( "501 Syntax error in IP6 address\r\n" );
      args = e;
      if( GetNum( &args, delim, &ip6.sin6_port ) == -1 )
         return SendReply( emsgClntInvPort );
      ip6.sin6_port = htons( ip6.sin6_port );
      break;
   default:
      return SendReply( "522 Unsupported address family. Use (1,2)\r\n" );
   }
   if( *args )
      return SendReply( "501 Extra data in address specification\r\n" );
   if( af == 1 )
      return ExecPort( (PSaddr)&ip4 );
   else
      return ExecPort( (PSaddr)&ip6 );
}

/*
* So-called passive mode: Client opens data connection to server.
*/

/* PASV command handler: Instructs the server to listen for incoming data
 * connection */
int Pasv( char* args )
{
   long addr;
   int h1, h2, h3, h4;
   short port;
   int p1, p2;

   if( EpsvOnly() )
      return SendReply( emsgNotImpl );
   if( ExtraParameters( args ) )
      return SendReplies();
   if( SrvrDataConn( AF_INET ) == RC_ERROR )
		return SendReplies();

   /* create data socket and force it to listen for incoming connections */
   if( CreateDataSocket( &hDataSocket) != RC_OK )
   {
      PError( 425 ); /* see comments for OpenDataConnection() */
      return SendReplies();
   }
   if( listen( hDataSocket, 1 ) < 0 )
   {
      close( hDataSocket );
      return SendReply( emsgPsvOpen );
   }
   bPassive = 1;

   /* Build the standard reply */
   addr = ntohl( ((struct sockaddr_in *)addrSrvrData)->sin_addr.s_addr );
   h1 = (addr >> 24) & 0xFF;
   h2 = (addr >> 16) & 0xFF;
   h3 = (addr >> 8) & 0xFF;
   h4 = addr & 0xFF;
   port = ntohs( ((struct sockaddr_in *)addrSrvrData)->sin_port );
   p1 = (port >> 8) & 0xFF;
   p2 = port & 0xFF;

   return SendReplyF
   ( 
      "227 Entering passive mode (%d,%d,%d,%d,%d,%d)\r\n", 
      h1, h2, h3, h4, p1, p2 
   );
}

int EPsv( char* args )
{
   sa_family_t af;
   
   if( !*args )
      af = addrSrvrCtrl->sa_family;
   else if( !strcasecmp( args, "all" ) )
   {
      af = addrSrvrCtrl->sa_family;
      bEpsvOnly = 1;
   }
   else if( !strcmp( args, "1" ) )
      af = AF_INET;
   else if( !strcmp( args, "2" ) )
      af = AF_INET6;
   else
      return SendReply( "501 Invalid parameter\r\n" );
   
	if( SrvrDataConn( af ) == RC_ERROR )
		return SendReplies();

   /* create data socket and force it to listen for incoming connections */
   if( CreateDataSocket( &hDataSocket) != RC_OK )
   {
      PError( 425 ); /* see comments for OpenDataConnection() */
      return SendReplies();
   }
   if( listen( hDataSocket, 1 ) < 0 )
   {
      close( hDataSocket );
      return SendReply( emsgPsvOpen );
   }
   bPassive = 1;

   return SendReplyF
   ( 
      "229 Entering extended passive mode (|||%u|)\r\n", 
      ntohs( SockAddrPort( addrSrvrData ) )
   );
}
