/*
   site.c

   Site specific SITE command.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: site.c,v 1.10 2008/04/30 23:14:20 shesmax Exp $ */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
//#include <time.h>
#include <errno.h>

#include <stdlib.h>
#include <string.h>

#include "libra.h"
#include "pi.h"
#include "dtp.h"
#include "parser.h"
#include "retcodes.h"
#include "sockets.h"
#include "listdir.h"
#include "log.h"
#include "xbuf.h"

static char sInvOctalNumber[] = 
   "502 \'%s\' should be an octal number from 000 to 777\r\n";
	
/* 
   SITE IDLE [<sp> timeout]
   
   Sets the timeout value 
*/
static int SiteIdle( char* args )
{
   unsigned long nValue;
   char* endp;

   if( *args )
   {
      nValue = strtoul( args, &endp, 0 );
      if( *endp || !SetSocketTimeout( nValue ) )
         return SendReplyF
         (
            "501 The timeout value should be a number "
            "in the range 0 ... %d\r\n", DEFAULT_TIMEOUT
         );
   }
   return SendReplyF( "200 Timeout: %ld secs\r\n", GetSocketTimeout() );
}

/* 
   SITE UMASK [<sp> umask]
   
   Changes umask.
*/
static int SiteUmask( char* args )
{
   unsigned long mask;
   char* endp;

   if( *args )
   {
      if( ReadOnlyBehavior() )
         return SendReply( "550 Access denied\r\n" );
      mask = strtoul( args, &endp, 8 );
      if( *endp || mask > 0777 )
         return SendReplyF( sInvOctalNumber, "umask" );
      umask( mask );
      LogEvent( LLOG_AUDIT, "Site Umask: %lo\n", mask );
   }
   mask = umask( 0 );   /* get the current mask */
   umask( mask );       /* Restore it */
   return SendReplyF( "200 %s%lo\r\n", mask <= 077 ? "0" : "", mask );
}

/* 
   SITE CHMOD <sp>mode <sp> file
   
   Changes the mode of the file. 
         
   The codes 450 and 550 are not specified in RFC 959 as a valid return
   codes for SITE. But the values mentioned there did not match to the
   purpose of this command. Hence I return 450 and 550 when appropriate.
*/
static int SiteChmod( char* args )
{
   unsigned long mode;
   char* p;

   if( ReadOnlyBehavior() )
      return SendReply( "550 Access denied\r\n" );
   
   mode = strtoul( args, &p, 8 );
   /* 
      I check for '\0' here because I wont to  complain 'Need path...'
      for incomplete commands like this one: site chmod 666
   */      
   if( p == args || (*p != ' ' && *p != '\0') || mode >  0777 )
      AccumReplyF( sInvOctalNumber, "mode" );
   else
   {
      args = *p == '\0' ? p : p + 1;
      if( GetPath( &args, &p, GP_OBLIGATORY ) )
         if( chmod( p, (mode_t)mode ) == -1 )
            PError( 450 );
         else
         {
            LogEvent( LLOG_AUDIT, "Site Chmod: %s %0o\n", p, mode );
            AccumReply( "200 File mode changed\r\n" );
         }
   }
   return SendReplies();
}

/* Callback function for SITE NEWER */
static int NewerListEntry
(
   const char* name, const struct stat* st, void* date
)
{
   char sPrefix[ 57 ];

   if( st->st_mtime > (time_t)date )
   {
      LsLongPrefix( st, sPrefix, sizeof sPrefix );
      AccumReply( sPrefix );
      AccumReplyPathname( name, strlen( name ) );
      return AccumReply( "\r\n" );
   }      
   return RC_OK;
}

/* 
   SITE NEWER <sp>  YYYYMMDDHHMMSS [<sp> pattern] 

   Lists all files that match to the given pattern (by
   default '*' is used) and that are newer then the given date.
   
   Note:
   
   WuFtpd opens a data connection to transfer the listing.
   I don't think that this is usable. This approach assumes that the client knows 
	it should open the data connection too. As you can imagine not all clients
   know that. The Berkeley FTP client and NcFTP supplied with RedHat are
   not prepared for this. Standard Windows NT FTP client fails to do that, etc.
   
   Although this command can be consumed by all clients (f.e. using QUOTE 
   command) if the listing  will be sent back through the control connection.
   Like STAT * that is part of RFC 959.  
*/
static int SiteNewer( char* args )
{
   time_t date = 0;
   char* sPath;

   if( !GetDate( &args, &date ) )
      return SendReplies();
   else if( !*args )
      sPath = "";
   else if( *args == ' ' )
   {
      ++args;
      if( !GetPath( &args, &sPath, GP_OBLIGATORY ) )
         return SendReplies();
   }
   else
      return SendReply( "501 Invalid format of argument(s)\r\n" );
   
	AccumReply( "200- " );
   AccumReplyPathname( sPath, strlen( sPath ) );
   AccumReply( "\r\n" );
   if( Ls( sPath, NewerListEntry, (void*)date ) == RC_OK )
      AccumReply( "200 OK\r\n" );
   else
      PError( 200 );
   return SendReplies();
}

static int SiteChecksum( char* args )
{
   if( *args )
   {
      if( strcasecmp( args, "on" ) == 0 )
         SetChecksumMode( RC_TRUE );
      else if( strcasecmp( args, "off" ) == 0 )
         SetChecksumMode( RC_FALSE );
      else
         return SendReply( "501 Invalid argument\r\n" );
   }
   return SendReplyF( "200 Checksum %s\r\n", ComputeChecksums() ? "on":"off" );
}

static int SiteHelp( char* unused UNUSED_PARAM )
{
   return Help( "SITE");
}

/* Please note that the array below should be sorted by keyword in ascending
 * order */
static Keyword site[] =
{
   {"CHECKSUM", &SiteChecksum, 0},
   {"CHMOD", &SiteChmod, 0},
   {"HELP", &SiteHelp, 0},
   {"IDLE", &SiteIdle, 0},
   {"MINFO", &SiteNewer, 0},
   {"NEWER", &SiteNewer, 0},
   {"UMASK", &SiteUmask, 0}
};
#define DIM( a ) (sizeof a / sizeof a[ 0 ])

int Site( char* sCmd )
{
   char* cmd = sCmd;
   char* args = strchr( cmd, ' ' );
   Keyword* pKwd;

   if( args )
      *args++ = '\0';   /* separate the command from its arguments */
   else
      args = cmd + strlen( cmd );

   pKwd = ResolveCommand( site, DIM( site ), cmd );

   if( pKwd )
      return pKwd->handler( args );
   return SendReply( "500 Unrecognized command\r\n" );
}
