/*
   sockets.c

   Wrappers for socket calls.
   struct sockaddr_... handling functions.

   Sockets are opened in non-blocking mode.
   Current implementation uses select() for timeouts.
   Timeouts are configurable by redefining of DEFAULT_TIMEOUT macro.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: sockets.c,v 1.20 2008/04/28 05:37:54 shesmax Exp $ */

#include <unistd.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/stat.h>      /* for struct stat */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sockets.h"
#include "inet6_xtoy.h"
#include "retcodes.h"
#include "dtp.h"           /* OOB data handler queries and controls DTP */

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0x4000
#endif

/*
* Timeouts
*/

static unsigned long nTimeout = DEFAULT_TIMEOUT;

unsigned long GetSocketTimeout()
{
   return nTimeout;
}

int SetSocketTimeout( unsigned long nTimeout_ )
{
   if( nTimeout_ > DEFAULT_TIMEOUT )
      return RC_FALSE;
   nTimeout = nTimeout_;
   return RC_TRUE;
}

/*
* Working with socket
*/

/* Waits for the data to be available for reading on socket nSocket */
static int Wait4DataAvail( int nSocket )
{
   fd_set fds;
   struct timeval tv = {nTimeout, 0};

   FD_ZERO( &fds );
   FD_SET( nSocket, &fds );
   return select( nSocket + 1, &fds, 0, 0, &tv ) == 1;
}

/* 
   Waits for the socket nSocket to be available for writing.
   
   Returns
   
   RC_OK          socket is available for writing
   RC_TIMEDOUT    timeout expired
   RC_ERROR       error on the socket.
*/
static int WaitOk2Write( int nSocket )
{
   fd_set fds;
   struct timeval tv = {nTimeout, 0};

   FD_ZERO( &fds );
   FD_SET( nSocket, &fds );
   if( select( nSocket + 1, 0, &fds, 0, &tv ) != -1 )
      return FD_ISSET( nSocket, &fds ) ? RC_OK : RC_TIMEDOUT;
   return RC_ERROR;
}

/* Sends the data to the remote host through the socket */
int WriteSocket( int nSocket, const void* pBuf, int nCount )
{
   struct msghdr mh;
   struct iovec iov;
   struct stat statbuf;
   int is_socket;
   int r;

   if( fstat( nSocket, &statbuf ) == -1)
      return RC_ERROR;
   if( S_ISFIFO( statbuf.st_mode ) )
      return RC_ERROR; /* we can't handle SIGPIPE for them */
   is_socket = 0;
   if( S_ISSOCK( statbuf.st_mode ) )
      is_socket = 1;

   iov.iov_base = (void *)pBuf;
   iov.iov_len = nCount;
   mh.msg_name = NULL;
   mh.msg_namelen = 0;
   mh.msg_iov = &iov;
   mh.msg_iovlen = 1;
   mh.msg_control = NULL;
   mh.msg_controllen = 0;
   mh.msg_flags = 0;
   while( iov.iov_len )
   {
      int nWritten;

      if( AbortTransfer() )
      {
         CancelTransferAbort();
         return RC_ABORTED;
      }
      
      r = WaitOk2Write( nSocket );
      if (r == RC_OK)
      {
         nWritten = is_socket
                     ? sendmsg( nSocket, &mh, MSG_NOSIGNAL )
                     : writev( nSocket, &iov, 1 );
         if( nWritten != -1 )
         {
            iov.iov_len -= nWritten;
            //*((char **)&iov.iov_base) += nWritten;
            iov.iov_base += nWritten;
         }
         else
            r = RC_ERROR;
      }
      switch( r )
      {
      case RC_TIMEDOUT:
         return RC_TIMEDOUT;
      case RC_ERROR:
         if( errno == EINTR )
            continue;
         return RC_ERROR;
      }
   }

   return RC_OK;
}

/* Reads not more than nCount bytes from the socket nSocket into the buffer */
int ReadSocket( int nSocket, void* pBuf, int nCount )
{
   if( AbortTransfer() )
   {
      CancelTransferAbort();
      return RC_ABORTED;
   }
   if( !Wait4DataAvail( nSocket ) )
      return RC_TIMEDOUT;
   return read( nSocket, pBuf, nCount );
}

/* Connects socket with the given address addr. */
int ConnectSocket( int nSocket, PSaddr addr )
{
   if( connect( nSocket, addr, SockAddrSize( addr ) ) != -1 )
      return RC_OK;
   if( errno != EINPROGRESS )
      return RC_ERROR;
   return WaitOk2Write( nSocket );
}

/* Accepts incoming connection requests on the given socket */
int AcceptSocket( int nSocket, PSaddr addr )
{
   socklen_t socklen = SockAddrSize( addr );
   int r;

   if( !Wait4DataAvail( nSocket ) )
      return INVALID_SOCKET;
   r = accept( nSocket, addr, &socklen );
   if( r == -1 )
       return r;
   if( socklen != SockAddrSize( addr ) )
   {
       close( r );
       return INVALID_SOCKET;
   }
   return r;
}

/* 
*  Socket addresses
*/

int AF_Supported( sa_family_t af )
{
   switch( af )
   {
   case AF_INET:
   case AF_INET6:
      return RC_TRUE;
   default:
      return RC_FALSE;
   }
}

char *AF_Name( sa_family_t af )
{
   switch( af )
   {
   case AF_INET:
      return "IPv4";
   case AF_INET6:
      return "IPv6"; 
   default:
      return "";
   }
}

PSaddr SockAddrCreate()
{
   return (PSaddr)malloc( sizeof( Saddr ) );
}

void SockAddrDestroy( PSaddr addr )
{
   if( addr )
      free( addr );
}

socklen_t SockAddrSize( PSaddr addr )
{
   if( !addr )
      return 0;
   switch( addr->sa_family )
   {
   case AF_INET:
      return sizeof( struct sockaddr_in );
   case AF_INET6:
      return sizeof( struct sockaddr_in6 );
   default:
      return 0;
   }
}

int SockAddrAssign( PSaddr dst, PSaddr src )
{
   if( !dst || !src )
      return RC_FALSE;
   memmove( dst, src, SockAddrSize( src ) );
   return RC_TRUE;
}

int SockAddrSetFamily( PSaddr addr, sa_family_t af )
{
   static char ipv4_compat[ 12 ] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
   static char ipv4_mapped[ 12 ] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF};
   typeof(((struct sockaddr_in *)addr)->sin_addr.s_addr) ipv4;
   typeof(((struct sockaddr_in *)addr)->sin_port) port;
   
#define IPV6_(a) ((struct sockaddr_in6 *)a)
#define IPV4_(a) ((struct sockaddr_in *)a)
   switch( addr->sa_family )
   {
   case AF_INET:
      if( af == AF_INET )
         return RC_TRUE;
      if( af != AF_INET6 )
         return RC_FALSE;    
      
      /* Build so-called IPv4-mapped IPv6 address. */
      ipv4 = IPV4_(addr)->sin_addr.s_addr;
      port = IPV4_(addr)->sin_port;
      addr->sa_family = AF_INET6;
      memset( IPV6_(addr)->sin6_addr.s6_addr, 0, sizeof IPV6_(0)->sin6_addr );
      IPV6_(addr)->sin6_addr.s6_addr32[ 3 ] = ipv4;
      IPV6_(addr)->sin6_addr.s6_addr16[ 5 ] = 0xFFFF;
      IPV6_(addr)->sin6_port = port;
      IPV6_(addr)->sin6_flowinfo = 0;
      return RC_TRUE;

   case AF_INET6:
      if( af == AF_INET6 )
         return RC_TRUE;
      if( af != AF_INET )
         return RC_FALSE;
      /* We can convert IPv4-compatible (::0:a.b.c.d) or IPv4-mapped
       * (::FFFF:a.b.c.d) IPv6 addresses. */ 
      if
      ( 
         memcmp( IPV6_(addr)->sin6_addr.s6_addr, ipv4_compat, sizeof ipv4_compat ) &&
         memcmp( IPV6_(addr)->sin6_addr.s6_addr, ipv4_mapped, sizeof ipv4_mapped )
      )
         return RC_FALSE;
      
      ipv4 = IPV6_(addr)->sin6_addr.s6_addr32[ 3 ];
      port = IPV6_(addr)->sin6_port;
      addr->sa_family = AF_INET;
      IPV4_(addr)->sin_addr.s_addr = ipv4;
      IPV4_(addr)->sin_port = port;
      return RC_TRUE;
   }
#undef IPV6_
#undef IPV4_
   return RC_FALSE;
}

int SockAddrSetPort( PSaddr addr, uint16_t port )
{
   if( !addr )
      return RC_FALSE;
   switch( addr->sa_family )
   {
   case AF_INET:
      ((Saddr *)addr)->ip4.sin_port = htons( port );
      return RC_TRUE;
   case AF_INET6:
      ((Saddr *)addr)->ip6.sin6_port = htons( port );
      return RC_TRUE;
   }
   return RC_FALSE;
}

uint16_t SockAddrPort( PSaddr addr )
{
   if( !addr )
      return 0;
   switch( addr->sa_family )
   {
   case AF_INET:
      return ((Saddr *)addr)->ip4.sin_port;
   case AF_INET6:
      return ((Saddr *)addr)->ip6.sin6_port;
   }
   return 0;
}

int SockAddrAny( PSaddr addr )
{
   if( !addr )
      return RC_FALSE;
   switch( addr->sa_family )
   {
   case AF_INET:
      ((Saddr *)addr)->ip4.sin_addr.s_addr = INADDR_ANY;
      ((Saddr *)addr)->ip4.sin_port = 0;
      break;
   case AF_INET6:
      ((Saddr *)addr)->ip6.sin6_addr = in6addr_any;
      ((Saddr *)addr)->ip6.sin6_port = 0;
      ((Saddr *)addr)->ip6.sin6_flowinfo = 0;
      break;
   }
   return RC_TRUE;
}

int SockAddrString( PSaddr addr, char *buf, size_t sz )
{
   if( !addr || !buf )
      return -1;
   switch( addr->sa_family )
   {
   case AF_INET:
      return snprintf
      (
         buf, sz, "%s port %d",
         inet_ntoa( ((struct sockaddr_in *)addr)->sin_addr ), 
         ntohs( ((struct sockaddr_in *)addr)->sin_port )
      );
   case AF_INET6:
   {
      char s[ 40 ];
      
      inet6_ntoa( &((struct sockaddr_in6 *)addr)->sin6_addr, s, sizeof s );
      return snprintf
      ( 
         buf, sz, "%s port %d", s, 
         ntohs( ((struct sockaddr_in6 *)addr)->sin6_port ) 
      );
   }
   }
   return -1;
}

