/*
   sockets.h

   Wrappers for socket calls

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: sockets.h,v 1.17 2008/04/28 05:37:54 shesmax Exp $ */

#ifndef __GUARD_SOCKETS_H
#define __GUARD_SOCKETS_H

#include "libra.h"

#define INVALID_SOCKET  -1

#include <sys/socket.h>
#include <netinet/in.h>
typedef union _Saddr
{
   struct sockaddr_in  ip4;
   struct sockaddr_in6 ip6;
} Saddr;
typedef struct sockaddr* PSaddr;       /* shorter name */

/*
 *  Socket operations
 */

/* 
 * Writes the given data to the socket nSocket.
 *  
 * Returns:
 * RC_OK          no errors
 * RC_ERROR       some error occured during write (see errno)
 * RC_ABORTED     data transfer aborted by the client (this can happen
 *                for data connections only)
 * RC_TIMEDOUT    operation timed out. 
 */
extern int WriteSocket( int nSocket, const void* pBuf, int nCount );

/* 
 * Reads the data from the socket nSocket into the buffer, but not more
 * then nCount bytes.
 * 
 * Returns:
 * number of bytes read so far if there were no errors.
 * RC_ERROR       some error occured during read (see errno)
 * RC_ABORTED     data transfer aborted by the client (this can happen
 *                for data connections only)
 * RC_TIMEDOUT    operation timed out. 
 */
extern int ReadSocket( int nSocket, void* pBuf, int nCount );

/* 
 * Connects the socket nSocket to the remote host whose address is in addr.
 * 
 * Returns:
 * RC_OK          no errors
 * RC_ERROR       OS error
 * RC_TIMEDOUT    operation timed out 
 */
extern int ConnectSocket( int nSocket, PSaddr addr );

/* 
 * Accepts the connection request on the socket nSocket.
 * Returns either socket handle or INVALID_SOCKET. 
 */
extern int AcceptSocket( int nSocket, PSaddr addr );

/*
 * Timeout values
 */

#ifndef _SYS_TYPES_H
#  include <sys/types.h>
#endif

/* Timeout configuration through recompilation only */
#define DEFAULT_TIMEOUT    300         /* 5 minutes (RFC 1123) */

/* Returns the timeout value for socket operation */
unsigned long GetSocketTimeout();

/* 
 * Sets a new timeout value. This should be less then DEFAULT_TIMEOUT.
 * Returns RC_TRUE(1) if a new value is set, otherwise returns RC_FALSE(0) 
 */
int SetSocketTimeout( unsigned long nNewTimeout );

/* 
 *  Socket address
 */

/* Returns RC_TRUE if the family af is supported by server */
int AF_Supported( sa_family_t af );

/* 
 * Returns the text name of address family. It the given address family is not
 * supported returns an empty string 
 */
char *AF_Name( sa_family_t af );

/* Creates memory block for PSaddr. Returns 0 if memory is exhausted. */
PSaddr SockAddrCreate();

/* Frees memory block allocated by SockAddrCreate() */
void SockAddrDestroy();

/* 
 * Returns the size of memory block occupied by addr. For unsupported address
 * families returns 0. 
 */
socklen_t SockAddrSize( PSaddr addr );

/* dst = src for PSaddr, Returns RC_TRUE if address was copied. */
int SockAddrAssign( PSaddr dst, PSaddr src );

/* 
 * Converts address dst so that it is a valid address of familiy af. 
 * 
 * Following conversions are possible:
 * 1. IPv4 -> IPv4-mapped IPv6         127.0.0.1         -> ::FFFF:127.0.0.1
 * 2. IPv4-compatible IPv6 -> IPv4     ::0:127.0.0.1     -> 127.0.0.1
 * 3. IPv4-mapped IPv6 -> IPv4         ::FFFF:127.0.0.1  -> 127.0.0.1
 * 
 * Returns RC_TRUE on success, RC_FALSE in case of unsupported conversion
 * request. 
 */
int SockAddrSetFamily( PSaddr addr, sa_family_t af );

/* 
 * Sets port value of addr to port. Returns RC_FALSE if addr belongs to
 * family that does not have ports. 
 */
int SockAddrSetPort( PSaddr addr, uint16_t port );

/* 
 * Returns port value of addr or -1 if addr belongs to family that 
 * does not have ports. 
 */
uint16_t SockAddrPort( PSaddr addr );

/* Sets address addr to so-called "address any" */
int SockAddrAny( PSaddr addr );

/* 
 * The  function  converts  the  Internet  address addr to a string that is
 * stored in a memory block that buf points to. sz defines the size of this
 * memory block. 
 */
int SockAddrString( PSaddr addr, char *buf, size_t sz );

#endif
