/*
   utf8.c

   Rudimentary implementation of routines dealing with UTF8 encoding.

   Copyright (C) 2002   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)
                 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/

/* $Id: utf8.c,v 1.5 2008/04/30 23:14:20 shesmax Exp $ */

#include "utf8.h"
#include "retcodes.h"

/*
 * Code is taken from RFC 2640 - Internationalization of the File Transfer Protocol
 *    Author: B. Curtin
 *    Date: July 1999
 *    Category: Proposed Standard
 */
           
int utf8_enctype(const unsigned char* buf, unsigned int len)
{
   const unsigned char* endbuf = buf + len;
   unsigned char byte2mask=0x00;
   unsigned char c;
   int trailing = 0;                   // trailing (continuation) bytes to follow
   int result = ENCTYPE_ASCII;

   while( buf != endbuf ) 
   {
      c = *buf++;
      if( trailing ) 
         if( (c & 0xC0) == 0x80 )      // Does trailing byte follow UTF-8 format?
         {
            if( byte2mask )            // Need to check 2nd byte for proper range?
               if (c & byte2mask)      // Are appropriate bits set?
                  byte2mask = 0x00;
               else
                  return ENCTYPE_LOCAL;
            trailing--; 
            result = ENCTYPE_UTF8;
         } 
         else
            return ENCTYPE_LOCAL;
      else if( (c & 0x80) == 0x00 )    // valid 1 byte UTF-8 
         continue;      
      else if( (c & 0xE0) == 0xC0 )    // valid 2 byte UTF-8
         if( !(c & 0x1F) )             // Is UTF-8 byte in proper range?
           return ENCTYPE_LOCAL;
         else
            trailing = 1;
      else if( (c&0xF0) == 0xE0 )      // valid 3 byte UTF-8
      {
         if(!(c & 0x0F) )              // Is UTF-8 byte in proper range?
            byte2mask = 0x20;          // If not set mask to check next byte
         trailing = 2;
      } 
      else if( (c&0xF8) == 0xF0 )      // valid 4 byte UTF-8
      {
         if( !(c & 0x07) )             // Is UTF-8 byte in proper range?
            byte2mask = 0x30;          // If not set mask to check next byte
         trailing = 3;
      } 
      else if( (c&0xFC) == 0xF8 )      // valid 5 byte UTF-8
      {
         if( !(c&0x03) )               // Is UTF-8 byte in proper range?
            byte2mask = 0x38;          // If not set mask to check next byte
         trailing = 4;
      } 
      else if( (c&0xFE) == 0xFC )      // valid 6 byte UTF-8
      {
         if( !(c&0x01) )               // Is UTF-8 byte in proper range?
            byte2mask = 0x3C;          // If not set mask to check next byte
         trailing = 5;
      } 
      else  
         return ENCTYPE_LOCAL; 
   }
   if (!trailing)
      return result;
   return ENCTYPE_LOCAL;
}

static unsigned int ucs4_to_utf8
( 
   const unsigned long* ucs4_buf, unsigned int ucs4_len, unsigned char* utf8_buf 
)
{
   const unsigned long* ucs4_endbuf = ucs4_buf + ucs4_len;
   unsigned int utf8_len = 0;             // return value for UTF-8 size
   unsigned char* t_utf8_buf = utf8_buf;  // Temporary pointer to load UTF-8 values

   while( ucs4_buf != ucs4_endbuf )
   {
      if( *ucs4_buf <= 0x7F )    // ASCII chars no conversion needed
      {
         *t_utf8_buf++ = (unsigned char) *ucs4_buf;
         utf8_len++;
         ucs4_buf++;
      }
      else if( *ucs4_buf <= 0x07FF ) // In the 2 byte utf-8 range
      {
         *t_utf8_buf++= (unsigned char) (0xC0 + (*ucs4_buf/0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + (*ucs4_buf%0x40));
         utf8_len+=2;
         ucs4_buf++;
      }
      else if ( *ucs4_buf <= 0xFFFF ) 
      /* In the 3 byte utf-8 range. The values 0x0000FFFE, 0x0000FFFF and 0x0000D800 - 0x0000DFFF do
         not occur in UCS-4 */
      {
         *t_utf8_buf++= (unsigned char) (0xE0 +
         (*ucs4_buf/0x1000));
         *t_utf8_buf++= (unsigned char) (0x80 +
         ((*ucs4_buf/0x40)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + (*ucs4_buf%0x40));
         utf8_len+=3;
         ucs4_buf++;
      }
      else if ( *ucs4_buf <= 0x1FFFFF ) //In the 4 byte utf-8 range
      {
         *t_utf8_buf++= (unsigned char) (0xF0 + (*ucs4_buf/0x040000));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x10000)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x40)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + (*ucs4_buf%0x40));
         utf8_len+=4;
         ucs4_buf++;

      }
      else if ( *ucs4_buf <= 0x03FFFFFF )//In the 5 byte utf-8 range
      {
         *t_utf8_buf++= (unsigned char) (0xF8 + (*ucs4_buf/0x01000000));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x040000)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x1000)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x40)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + (*ucs4_buf%0x40));
         utf8_len+=5;
         ucs4_buf++;
      }
      else if ( *ucs4_buf <= 0x7FFFFFFF )  /* In the 6 byte utf-8 range */
      {
         *t_utf8_buf++= (unsigned char) (0xF8 +(*ucs4_buf/0x40000000));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x01000000)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x040000)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x1000)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + ((*ucs4_buf/0x40)%0x40));
         *t_utf8_buf++= (unsigned char) (0x80 + (*ucs4_buf%0x40));
         utf8_len+=6;
         ucs4_buf++;
      }
   }
   return (utf8_len);
}

static unsigned int utf8_to_ucs4( const unsigned char* utf8, unsigned int utf8_len, unsigned long* ucs4 )
{
   const unsigned char* utf8_endbuf = utf8 + utf8_len;
   unsigned int ucs_len = 0;

   while( utf8 != utf8_endbuf )
   {
      if( (*utf8 & 0x80) == 0x00 )     /* ASCII chars no conversion is needed */
      {
         *ucs4++ = (unsigned long)*utf8;
         utf8++;
         ucs_len++;
      }
      else if( (*utf8 & 0xE0)== 0xC0 ) /* In the 2 byte utf-8 range */
      {
         *ucs4++ = (unsigned long)(((*utf8 - 0xC0) * 0x40) + (*(utf8 + 1) - 0x80));
         utf8 += 2;
         ucs_len++;
      }
      else if( (*utf8 & 0xF0) == 0xE0 ) /*In the 3 byte utf-8 range */
      {
         *ucs4++ = (unsigned long)(((*utf8 - 0xE0) * 0x1000) + ((*(utf8+1) -  0x80) * 0x40)
            + (*(utf8+2) - 0x80));
         utf8 += 3;
         ucs_len++;
      }
      else if( (*utf8 & 0xF8) == 0xF0 ) /* In the 4 byte utf-8 range */
      {
         *ucs4++ = (unsigned long)
            (((*utf8 - 0xF0) * 0x040000)
            + ((*(utf8 + 1) -  0x80) * 0x1000)
            + ((*(utf8 + 2) -  0x80) * 0x40)
            + (*(utf8 + 3) - 0x80));
         utf8 += 4;
         ucs_len++;
      }
      else if( (*utf8 & 0xFC) == 0xF8 ) /* In the 5 byte utf-8 range */
      {
         *ucs4++ = (unsigned long)
            (((*utf8 - 0xF8) * 0x01000000)
            + ((*(utf8 + 1) - 0x80) * 0x040000)
            + ((*(utf8 + 2) -  0x80) * 0x1000)
            + ((*(utf8 + 3) -  0x80) * 0x40)
            + (*(utf8 + 4) - 0x80));
         utf8 += 5;
         ucs_len++;
      }
      else if( (*utf8 & 0xFE) == 0xFC ) /* In the 6 byte utf-8 range */
      {
         *ucs4++ = (unsigned long)
            (((*utf8 - 0xFC) * 0x40000000)
            + ((*(utf8 + 1) - 0x80) * 0x010000000)
            + ((*(utf8 + 2) - 0x80) * 0x040000)
            + ((*(utf8 + 3) -  0x80) * 0x1000)
            + ((*(utf8 + 4) -  0x80) * 0x40)
            + (*(utf8 + 5) - 0x80));
         utf8 += 6;
         ucs_len++;
      }
   }
   return ucs_len;
}

static struct CvtTbl
{
   unsigned char local;
   unsigned long ucs4;
}
cvttab[ 256 ];
/*
{
	{0x00, 0x0000},   //  NULL
	{0x01, 0x0001},	//  START OF HEADING
	{0x02, 0x0002},	//  START OF TEXT
	{0x03, 0x0003},	//  END OF TEXT
	{0x04, 0x0004},	//  END OF TRANSMISSION
	{0x05, 0x0005},	//  ENQUIRY
	{0x06, 0x0006},	//  ACKNOWLEDGE
	{0x07, 0x0007},	//  BELL
	{0x08, 0x0008},	//  BACKSPACE
	{0x09, 0x0009},	//  HORIZONTAL TABULATION
	{0x0a, 0x000a},	//  LINE FEED
	{0x0b, 0x000b},	//  VERTICAL TABULATION
	{0x0c, 0x000c},	//  FORM FEED
	{0x0d, 0x000d},	//  CARRIAGE RETURN
	{0x0e, 0x000e},	//  SHIFT OUT
	{0x0f, 0x000f},	//  SHIFT IN
	{0x10, 0x0010},	//  DATA LINK ESCAPE
	{0x11, 0x0011},	//  DEVICE CONTROL ONE
	{0x12, 0x0012},	//  DEVICE CONTROL TWO
	{0x13, 0x0013},	//  DEVICE CONTROL THREE
	{0x14, 0x0014},	//  DEVICE CONTROL FOUR
	{0x15, 0x0015},	//  NEGATIVE ACKNOWLEDGE
	{0x16, 0x0016},	//  SYNCHRONOUS IDLE
	{0x17, 0x0017},	//  END OF TRANSMISSION BLOCK
	{0x18, 0x0018},	//  CANCEL
	{0x19, 0x0019},	//  END OF MEDIUM
	{0x1a, 0x001a},	//  SUBSTITUTE
	{0x1b, 0x001b},	//  ESCAPE
	{0x1c, 0x001c},	//  FILE SEPARATOR
	{0x1d, 0x001d},	//  GROUP SEPARATOR
	{0x1e, 0x001e},	//  RECORD SEPARATOR
	{0x1f, 0x001f},	//  UNIT SEPARATOR
	{0x20, 0x0020},	//  SPACE
	{0x21, 0x0021},	//  EXCLAMATION MARK
	{0x22, 0x0022},	//  QUOTATION MARK
	{0x23, 0x0023},	//  NUMBER SIGN
	{0x24, 0x0024},	//  DOLLAR SIGN
	{0x25, 0x0025},	//  PERCENT SIGN
	{0x26, 0x0026},	//  AMPERSAND
	{0x27, 0x0027},	//  APOSTROPHE
	{0x28, 0x0028},	//  LEFT PARENTHESIS
	{0x29, 0x0029},	//  RIGHT PARENTHESIS
	{0x2a, 0x002a},	//  ASTERISK
	{0x2b, 0x002b},	//  PLUS SIGN
	{0x2c, 0x002c},	//  COMMA
	{0x2d, 0x002d},	//  HYPHEN-MINUS
	{0x2e, 0x002e},	//  FULL STOP
	{0x2f, 0x002f},	//  SOLIDUS
	{0x30, 0x0030},	//  DIGIT ZERO
	{0x31, 0x0031},	//  DIGIT ONE
	{0x32, 0x0032},	//  DIGIT TWO
	{0x33, 0x0033},	//  DIGIT THREE
	{0x34, 0x0034},	//  DIGIT FOUR
	{0x35, 0x0035},	//  DIGIT FIVE
	{0x36, 0x0036},	//  DIGIT SIX
	{0x37, 0x0037},	//  DIGIT SEVEN
	{0x38, 0x0038},	//  DIGIT EIGHT
	{0x39, 0x0039},	//  DIGIT NINE
	{0x3a, 0x003a},	//  COLON
	{0x3b, 0x003b},	//  SEMICOLON
	{0x3c, 0x003c},	//  LESS-THAN SIGN
	{0x3d, 0x003d},	//  EQUALS SIGN
	{0x3e, 0x003e},	//  GREATER-THAN SIGN
	{0x3f, 0x003f},	//  QUESTION MARK
	{0x40, 0x0040},	//  COMMERCIAL AT
	{0x41, 0x0041},	//  LATIN CAPITAL LETTER A
	{0x42, 0x0042},	//  LATIN CAPITAL LETTER B
	{0x43, 0x0043},	//  LATIN CAPITAL LETTER C
	{0x44, 0x0044},	//  LATIN CAPITAL LETTER D
	{0x45, 0x0045},	//  LATIN CAPITAL LETTER E
	{0x46, 0x0046},	//  LATIN CAPITAL LETTER F
	{0x47, 0x0047},	//  LATIN CAPITAL LETTER G
	{0x48, 0x0048},	//  LATIN CAPITAL LETTER H
	{0x49, 0x0049},	//  LATIN CAPITAL LETTER I
	{0x4a, 0x004a},	//  LATIN CAPITAL LETTER J
	{0x4b, 0x004b},	//  LATIN CAPITAL LETTER K
	{0x4c, 0x004c},	//  LATIN CAPITAL LETTER L
	{0x4d, 0x004d},	//  LATIN CAPITAL LETTER M
	{0x4e, 0x004e},	//  LATIN CAPITAL LETTER N
	{0x4f, 0x004f},	//  LATIN CAPITAL LETTER O
	{0x50, 0x0050},	//  LATIN CAPITAL LETTER P
	{0x51, 0x0051},	//  LATIN CAPITAL LETTER Q
	{0x52, 0x0052},	//  LATIN CAPITAL LETTER R
	{0x53, 0x0053},	//  LATIN CAPITAL LETTER S
	{0x54, 0x0054},	//  LATIN CAPITAL LETTER T
	{0x55, 0x0055},	//  LATIN CAPITAL LETTER U
	{0x56, 0x0056},	//  LATIN CAPITAL LETTER V
	{0x57, 0x0057},	//  LATIN CAPITAL LETTER W
	{0x58, 0x0058},	//  LATIN CAPITAL LETTER X
	{0x59, 0x0059},	//  LATIN CAPITAL LETTER Y
	{0x5a, 0x005a},	//  LATIN CAPITAL LETTER Z
	{0x5b, 0x005b},	//  LEFT SQUARE BRACKET
	{0x5c, 0x005c},	//  REVERSE SOLIDUS
	{0x5d, 0x005d},	//  RIGHT SQUARE BRACKET
	{0x5e, 0x005e},	//  CIRCUMFLEX ACCENT
	{0x5f, 0x005f},	//  LOW LINE
	{0x60, 0x0060},	//  GRAVE ACCENT
	{0x61, 0x0061},	//  LATIN SMALL LETTER A
	{0x62, 0x0062},	//  LATIN SMALL LETTER B
	{0x63, 0x0063},	//  LATIN SMALL LETTER C
	{0x64, 0x0064},	//  LATIN SMALL LETTER D
	{0x65, 0x0065},	//  LATIN SMALL LETTER E
	{0x66, 0x0066},	//  LATIN SMALL LETTER F
	{0x67, 0x0067},	//  LATIN SMALL LETTER G
	{0x68, 0x0068},	//  LATIN SMALL LETTER H
	{0x69, 0x0069},	//  LATIN SMALL LETTER I
	{0x6a, 0x006a},	//  LATIN SMALL LETTER J
	{0x6b, 0x006b},	//  LATIN SMALL LETTER K
	{0x6c, 0x006c},	//  LATIN SMALL LETTER L
	{0x6d, 0x006d},	//  LATIN SMALL LETTER M
	{0x6e, 0x006e},	//  LATIN SMALL LETTER N
	{0x6f, 0x006f},	//  LATIN SMALL LETTER O
	{0x70, 0x0070},	//  LATIN SMALL LETTER P
	{0x71, 0x0071},	//  LATIN SMALL LETTER Q
	{0x72, 0x0072},	//  LATIN SMALL LETTER R
	{0x73, 0x0073},	//  LATIN SMALL LETTER S
	{0x74, 0x0074},	//  LATIN SMALL LETTER T
	{0x75, 0x0075},	//  LATIN SMALL LETTER U
	{0x76, 0x0076},	//  LATIN SMALL LETTER V
	{0x77, 0x0077},	//  LATIN SMALL LETTER W
	{0x78, 0x0078},	//  LATIN SMALL LETTER X
	{0x79, 0x0079},	//  LATIN SMALL LETTER Y
	{0x7a, 0x007a},	//  LATIN SMALL LETTER Z
	{0x7b, 0x007b},	//  LEFT CURLY BRACKET
	{0x7c, 0x007c},	//  VERTICAL LINE
	{0x7d, 0x007d},	//  RIGHT CURLY BRACKET
	{0x7e, 0x007e},	//  TILDE
	{0x7f, 0x007f},	//  DELETE
	{0x80, 0x2500},	//  BOX DRAWINGS LIGHT HORIZONTAL
	{0x81, 0x2502},	//  BOX DRAWINGS LIGHT VERTICAL
	{0x82, 0x250c},	//  BOX DRAWINGS LIGHT DOWN AND RIGHT
	{0x83, 0x2510},	//  BOX DRAWINGS LIGHT DOWN AND LEFT
	{0x84, 0x2514},	//  BOX DRAWINGS LIGHT UP AND RIGHT
	{0x85, 0x2518},	//  BOX DRAWINGS LIGHT UP AND LEFT
	{0x86, 0x251c},	//  BOX DRAWINGS LIGHT VERTICAL AND RIGHT
	{0x87, 0x2524},	//  BOX DRAWINGS LIGHT VERTICAL AND LEFT
	{0x88, 0x252c},	//  BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
	{0x89, 0x2534},	//  BOX DRAWINGS LIGHT UP AND HORIZONTAL
	{0x8a, 0x253c},	//  BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
	{0x8b, 0x2580},	//  UPPER HALF BLOCK
	{0x8c, 0x2584},	//  LOWER HALF BLOCK
	{0x8d, 0x2588},	//  FULL BLOCK
	{0x8e, 0x258c},	//  LEFT HALF BLOCK
	{0x8f, 0x2590},	//  RIGHT HALF BLOCK
	{0x90, 0x2591},	//  LIGHT SHADE
	{0x91, 0x2592},	//  MEDIUM SHADE
	{0x92, 0x2593},	//  DARK SHADE
	{0x93, 0x2320},	//  UPPER HALF OF INTEGRAL
	{0x94, 0x25a0},	//  BLACK SQUARE
	{0x95, 0x2219},	//  BULLET OPERATOR
	{0x96, 0x221a},	//  SQUARE ROOT
	{0x97, 0x2248},	//   ALMOST EQUAL TO
	{0x98, 0x2264},	//  LESS-THAN OR EQUAL TO
	{0x99, 0x2265},	//  GREATER-THAN OR EQUAL TO
	{0x9a, 0x00a0},	//  NO-BREAK SPACE
	{0x9b, 0x2321},	//  LOWER HALF OF INTEGRAL
	{0x9c, 0x00b0},	//  DEGREE SIGN
	{0x9d, 0x00b2},	//  SUPERSCRIPT TWO
	{0x9e, 0x00b7},	//  MIDDLE DOT
	{0x9f, 0x00f7},	//  DIVISION SIGN 
	{0xa0, 0x2550},	//  BOX DRAWINGS DOUBLE HORIZONTAL
	{0xa1, 0x2551},	//  BOX DRAWINGS DOUBLE VERTICAL
	{0xa2, 0x2552},	//  BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE
	{0xa3, 0x0451},	//  CYRILLIC SMALL LETTER IO
	{0xa4, 0x2553},	//  BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE
	{0xa5, 0x2554},	//  BOX DRAWINGS DOUBLE DOWN AND RIGHT
	{0xa6, 0x2555},	//  BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE
	{0xa7, 0x2556},	//  BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE
	{0xa8, 0x2557},	//  BOX DRAWINGS DOUBLE DOWN AND LEFT
	{0xa9, 0x2558},	//  BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE
	{0xaa, 0x2559},	//  BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE
	{0xab, 0x255a},	//  BOX DRAWINGS DOUBLE UP AND RIGHT
	{0xac, 0x255b},	//  BOX DRAWINGS UP SINGLE AND LEFT DOUBLE
	{0xad, 0x255c},	//  BOX DRAWINGS UP DOUBLE AND LEFT SINGLE
	{0xae, 0x255d},	//  BOX DRAWINGS DOUBLE UP AND LEFT
	{0xaf, 0x255e},	//  BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE
	{0xb0, 0x255f},	//  BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE
	{0xb1, 0x2560},	//  BOX DRAWINGS DOUBLE VERTICAL AND RIGHT
	{0xb2, 0x2561},	//  BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE
	{0xb3, 0x0401},	//  CYRILLIC CAPITAL LETTER IO
	{0xb4, 0x2562},	//  BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE
	{0xb5, 0x2563},	//  BOX DRAWINGS DOUBLE VERTICAL AND LEFT
	{0xb6, 0x2564},	//  BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE
	{0xb7, 0x2565},	//  BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE
	{0xb8, 0x2566},	//  BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL
	{0xb9, 0x2567},	//  BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE
	{0xba, 0x2568},	//  BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE
	{0xbb, 0x2569},	//  BOX DRAWINGS DOUBLE UP AND HORIZONTAL
	{0xbc, 0x256a},	//  BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE
	{0xbd, 0x256b},	//  BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE
	{0xbe, 0x256c},	//  BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL
	{0xbf, 0x00a9},	//  COPYRIGHT SIGN 
	{0xc0, 0x044e},	//  CYRILLIC SMALL LETTER YU
	{0xc1, 0x0430},	//  CYRILLIC SMALL LETTER A
	{0xc2, 0x0431},	//  CYRILLIC SMALL LETTER BE
	{0xc3, 0x0446},	//  CYRILLIC SMALL LETTER TSE
	{0xc4, 0x0434},	//  CYRILLIC SMALL LETTER DE
	{0xc5, 0x0435},	//  CYRILLIC SMALL LETTER IE
	{0xc6, 0x0444},	//  CYRILLIC SMALL LETTER EF
	{0xc7, 0x0433},	//  CYRILLIC SMALL LETTER GHE
	{0xc8, 0x0445},	//  CYRILLIC SMALL LETTER HA
	{0xc9, 0x0438},	//  CYRILLIC SMALL LETTER I
	{0xca, 0x0439},	//  CYRILLIC SMALL LETTER SHORT I
	{0xcb, 0x043a},	//  CYRILLIC SMALL LETTER KA
	{0xcc, 0x043b},	//  CYRILLIC SMALL LETTER EL
	{0xcd, 0x043c},	//  CYRILLIC SMALL LETTER EM
	{0xce, 0x043d},	//  CYRILLIC SMALL LETTER EN
	{0xcf, 0x043e},	//  CYRILLIC SMALL LETTER O
	{0xd0, 0x043f},	//  CYRILLIC SMALL LETTER PE
	{0xd1, 0x044f},	//  CYRILLIC SMALL LETTER YA
	{0xd2, 0x0440},	//  CYRILLIC SMALL LETTER ER
	{0xd3, 0x0441},	//  CYRILLIC SMALL LETTER ES
	{0xd4, 0x0442},	//  CYRILLIC SMALL LETTER TE
	{0xd5, 0x0443},	//  CYRILLIC SMALL LETTER U
	{0xd6, 0x0436},	//  CYRILLIC SMALL LETTER ZHE
	{0xd7, 0x0432},	//  CYRILLIC SMALL LETTER VE
	{0xd8, 0x044c},	//  CYRILLIC SMALL LETTER SOFT SIGN
	{0xd9, 0x044b},	//  CYRILLIC SMALL LETTER YERU
	{0xda, 0x0437},	//  CYRILLIC SMALL LETTER ZE
	{0xdb, 0x0448},	//  CYRILLIC SMALL LETTER SHA
	{0xdc, 0x044d},	//  CYRILLIC SMALL LETTER E
	{0xdd, 0x0449},	//  CYRILLIC SMALL LETTER SHCHA
	{0xde, 0x0447},	//  CYRILLIC SMALL LETTER CHE
	{0xdf, 0x044a},	//  CYRILLIC SMALL LETTER HARD SIGN
	{0xe0, 0x042e},	//  CYRILLIC CAPITAL LETTER YU
	{0xe1, 0x0410},	//  CYRILLIC CAPITAL LETTER A
	{0xe2, 0x0411},	//  CYRILLIC CAPITAL LETTER BE
	{0xe3, 0x0426},	//  CYRILLIC CAPITAL LETTER TSE
	{0xe4, 0x0414},	//  CYRILLIC CAPITAL LETTER DE
	{0xe5, 0x0415},	//  CYRILLIC CAPITAL LETTER IE
	{0xe6, 0x0424},	//  CYRILLIC CAPITAL LETTER EF
	{0xe7, 0x0413},	//  CYRILLIC CAPITAL LETTER GHE
	{0xe8, 0x0425},	//  CYRILLIC CAPITAL LETTER HA
	{0xe9, 0x0418},	//  CYRILLIC CAPITAL LETTER I
	{0xea, 0x0419},	//  CYRILLIC CAPITAL LETTER SHORT I
	{0xeb, 0x041a},	//  CYRILLIC CAPITAL LETTER KA
	{0xec, 0x041b},	//  CYRILLIC CAPITAL LETTER EL
	{0xed, 0x041c},	//  CYRILLIC CAPITAL LETTER EM
	{0xee, 0x041d},	//  CYRILLIC CAPITAL LETTER EN
	{0xef, 0x041e},	//  CYRILLIC CAPITAL LETTER O
	{0xf0, 0x041f},	//  CYRILLIC CAPITAL LETTER PE
	{0xf1, 0x042f},	//  CYRILLIC CAPITAL LETTER YA
	{0xf2, 0x0420},	//  CYRILLIC CAPITAL LETTER ER
	{0xf3, 0x0421},	//  CYRILLIC CAPITAL LETTER ES
	{0xf4, 0x0422},	//  CYRILLIC CAPITAL LETTER TE
	{0xf5, 0x0423},	//  CYRILLIC CAPITAL LETTER U
	{0xf6, 0x0416},	//  CYRILLIC CAPITAL LETTER ZHE
	{0xf7, 0x0412},	//  CYRILLIC CAPITAL LETTER VE
	{0xf8, 0x042c},	//  CYRILLIC CAPITAL LETTER SOFT SIGN
	{0xf9, 0x042b},	//  CYRILLIC CAPITAL LETTER YERU
	{0xfa, 0x0417},	//  CYRILLIC CAPITAL LETTER ZE
	{0xfb, 0x0428},	//  CYRILLIC CAPITAL LETTER SHA
	{0xfc, 0x042d},	//  CYRILLIC CAPITAL LETTER E
	{0xfd, 0x0429},	//  CYRILLIC CAPITAL LETTER SHCHA
	{0xfe, 0x0427},	//  CYRILLIC CAPITAL LETTER CHE
	{0xff, 0x042a} 	//  CYRILLIC CAPITAL LETTER HARD SIGN
};
*/
size_t cvttab_len = sizeof cvttab / sizeof cvttab[0];

#include <stdio.h>
size_t utf8_to_local( const unsigned char *utf8, size_t len, unsigned char **local )
{
   unsigned long* ucs4 = (unsigned long*)malloc( len * sizeof (unsigned long) );
   unsigned int i;
   unsigned int j;

   if( !ucs4 )
      return 0;
   len = utf8_to_ucs4( utf8, len, ucs4 );  // len after <= len before
   if( !(*local = (unsigned char*)malloc( len + 1 )) )
   {
      free( ucs4 );
      return 0;
   }
   for (i = 0; i < len; ++i)
   {
      j = 0;
      (*local)[ i ] = '_';

      while( j < cvttab_len )
      {
         if(cvttab[ j ].ucs4 == ucs4[ i ] )
         {
            (*local)[ i ] = cvttab[ j ].local;
            break;
         }
         ++j;
      }
   }
   free( ucs4 );
   (*local)[ len ] = '\0';
   return len;
}

size_t local_to_utf8( const unsigned char *local, size_t len, unsigned char **utf8 )
{
   unsigned long* ucs4 = (unsigned long*)malloc( (len + 1) * sizeof (unsigned long) );
   unsigned int i;
   unsigned int j;

   if( !ucs4 )
      return 0;
   if( !(*utf8 = (unsigned char*)malloc( 6 * len + 1 )) )
   {
      free( ucs4 );
      return 0;
   }
   
   // Convert local to UCS-4
   for (i = 0; i < len; ++i)
   {
      j = 0;
      ucs4[ i ] = '_';
      
      while( j < cvttab_len )
      {
         if(cvttab[ j ].local == local[ i ] )
         {
            ucs4[ i ] = cvttab[ j ].ucs4;
            break;
         }
         ++j;
      }
   }
   ucs4[ i ] = 0;
   // Convert UCS-4 to UTF-8
   len = ucs4_to_utf8( ucs4, len + 1, *utf8 );  // len after >= len before
   
   free( ucs4 );
   return len;
}

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdio.h>
#include <string.h>
#include <ctype.h>
            
// Helper routine for utf8_read_local_to_unicode. Processes one line.
int ParseNumbers( char *line )
{
   unsigned long local;
   unsigned long ucs4;
   char* endp;

   local = strtoul( line, &endp, 16 );
   if( endp == line || local > 256 )
      return RC_ERROR;
   line = endp;
   ucs4 = strtoul( line, &endp, 16 );
   if( endp == line ) 
      return RC_ERROR;
            
   if( cvttab_len > 256 )
      return RC_ERROR;
   cvttab[ cvttab_len ].local = local;
   cvttab[ cvttab_len ].ucs4 = ucs4;
   ++cvttab_len;
   return RC_OK;
}


int utf8_read_local_to_unicode( const char* file )
{
   int fd = open( file, O_RDONLY );
   struct stat st;
   const char* buf;
   int result;

   cvttab_len = 0;
   
   if( fd == -1 )
      return RC_ERROR;
      
   fstat( fd, &st );
   buf = mmap( 0, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0 );
   if( buf != (void*) -1 )
   {
      const char* begin = buf;
      const char* end = buf + st.st_size;

      result = RC_OK;
      while( begin < end )
      {
         const char* eol = memchr( begin, '\n', end - begin );
         const char *next;
         char* line;
         char* lend;
         size_t sz;
         
         // Get the line from file
         if( !eol )
            eol = end;
         next = eol + 1;
         sz = eol - begin;
         if( !(line = (char *)malloc( sz + 1 )) )
         {
            result = RC_ERROR;
            break;
         }
         memmove( line, begin, sz );
         line[ sz ] = '\0';

         // Remove white spaces (including comments) at the end of line
         if( (lend = memchr( line, '#', sz )) )
            *lend = '\0';
         else 
            lend = line + sz + 1; 
         while( 1 )
            if( --lend < line || (isascii( *lend ) && !isspace( *lend )) )
               break;
         *++lend = '\0';
         
         result = *line ? ParseNumbers( line ) : RC_OK;
         
         // Advance to next line
         free( line );
         if( result != RC_OK )
            break;
         begin = next;
      }  
      munmap( (char *)buf, st.st_size );
   }
   else
      result = RC_ERROR;

   close( fd );
   return result;
}        
