/*
   xbuf.h

   XBuf allows to create either a buffer of fixed length or an expandable
   buffer. The size of the expandable buffer is limited only be the
   size of the heap.

   The implementation provides several functions allow to append
   the desired text at the end of the buffer.

   This buffer is heavily used in the program for different purposes.

   Copyright (C) 1999   Andrey Savochkin (saw@saw.sw.com.sg)
                        Maxim Shesterikov (libraftp@narod.ru)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA  02111-1307, USA.
*/
/* $Id: xbuf.h,v 1.13 2008/04/28 05:37:54 shesmax Exp $ */

#ifndef __GUARD_XBUF_H
#define __GUARD_XBUF_H

#include <sys/types.h>  /* for size_t */

typedef struct tagXBuffer
{
	char*  data;       /* points to memory allocated for the buffer */
	size_t nUsed;      /* the length of the text in the buffer */
	size_t nSize;      /* the current size of the buffer */
	int    bFixed;     /* 1 means that the buffer cannot grow */
} XBuf;

/* Return codes for XBuf routines */
#define XBUF_BADPARAM           -2
#define XBUF_OVERFLOW           -1
#define XBUF_OK                  0

/*
*  Creation / destruction
*/

/* 
	Creates the fixed buffer of the desired size. Returns either XBUF_OK,
   or XBUF_OVERFLOW if there is not enough heap for buffer. 
*/
int CreateFixedXBuf( XBuf* pBuf, size_t nSize );
/* 
	Creates the expandable buffer of the desired size. Returns either XBUF_OK,
   or XBUF_OVERFLOW if there is not enough heap for buffer. 
*/
int CreateExpandableXBuf( XBuf* pBuf, size_t nSize );
/* Frees the memory allocated for the buffer */
void DestroyXBuf( XBuf* pBuf );

/*
*   Output to the buffer 
*/

#ifndef __STDARG_H
#	include <stdarg.h>
#endif

/* 
	Appends the given text to the end of the buffer. Returns the number
   of bytes written to the buffer.
   If the given text does not fit into the buffer:
   -  If the buffer is fixed and there are no free space left
      returns XBUF_OVERFLOW, otherwise writes as many bytes as there is
      free space.
   -  If the buffer is expandable tries to increase its size. If this fails
      returns XBUF_OVERFLOW, otherwise returns nBytes
*/
int WriteXBuf( XBuf* pBuf, const char* data, size_t nBytes );
/* 
	Appends the formatted text to the end of the buffer. The syntax
   of the format specifier fmt is the same as for printf.
   Returns the number of bytes written to the buffer if all is okay.
   If the given text does not fit into the buffer:
   -  If the buffer has fixed length returns XBUF_OVERFLOW.
   -  If the buffer is expandable increases the buffer to the size
      big enough to include the formatted text. If there is not
      enough heap for such a big buffer returns XBUF_OVERFLOW.
*/
int VPrintfXBuf( XBuf* pBuf, const char* fmt, va_list args );
/* Wrapper for VPrintfXBuf */
int PrintfXBuf( XBuf* pBuf, const char* fmt, ... );

/*
* Buffers & sockets
*/

/* 
  Flushes the content of the Xbuf to the given socket. Returns RC_OK
  on success, RC_ERROR etc. in case of error. 
*/
int FlushSockXBuf( XBuf* pXBuf, int hSocket );

/* 
  Appends the data that is pointed by buf to the end of the XBuf pXBuf.
  Appends as much as possible. If there is not enough space flushes pXBuf 
  to the socket hSocket by calling Flush, then appends next portion of buf
  and so on. 
  Returns RC_OK on success, otherwise returns RC_ERROR or other error codes.
*/
int PutSockXBuf
( 
   XBuf* pXBuf, const char* buf, unsigned nSize, 
   int (*Flush)( XBuf* pXBuf, int socket ),  int socket 
);

/* 
   Reads data from the socket, but not more then the number of unused bytes in
   the buffer. Returns number of bytes read so far. 
*/
int ReadSockXBuf( XBuf* xb, int socket );

/*
*  Miscellaneous
*/

size_t SizeOfXBuf( XBuf* xb );
size_t UsedSpaceXBuf( XBuf* xb );
size_t FreeSpaceXBuf( XBuf* xb );

/* Removes len bytes from the beginning of the buffer. */
void LTrimXBuf( XBuf* xb, unsigned len );

/* Clears the content of the buffer */
void ClearXBuf( XBuf* pBuf );

/*
* Checks if the buffer contains line terminator (\r\n) and returns the line
* length including the terminator.
* If no line terminator is found return 0.
*/
unsigned LineLenXBuf( XBuf* xb );

#endif
